#
# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tests for USEquityPricingLoader and related classes.
"""
from nose_parameterized import parameterized
from numpy import (
    arange,
    datetime64,
    float64,
    ones,
    uint32,
)
from numpy.testing import (
    assert_allclose,
    assert_array_equal,
)
from pandas import (
    concat,
    DataFrame,
    Int64Index,
    Timestamp,
)
from pandas.util.testing import assert_frame_equal
from toolz.curried.operator import getitem

from zipline.lib.adjustment import Float64Multiply
from zipline.pipeline.domain import US_EQUITIES
from zipline.pipeline.loaders.synthetic import (
    NullAdjustmentReader,
    make_bar_data,
    expected_bar_values_2d,
)
from zipline.pipeline.loaders.equity_pricing_loader import (
    USEquityPricingLoader,
)

from zipline.errors import WindowLengthTooLong
from zipline.pipeline.data import USEquityPricing
from zipline.testing import (
    seconds_to_timestamp,
    str_to_seconds,
    MockDailyBarReader,
)
from zipline.testing.fixtures import (
    WithAdjustmentReader,
    ZiplineTestCase,
)

# Test calendar ranges over the month of June 2015
#      June 2015
# Mo Tu We Th Fr Sa Su
#  1  2  3  4  5  6  7
#  8  9 10 11 12 13 14
# 15 16 17 18 19 20 21
# 22 23 24 25 26 27 28
# 29 30
TEST_CALENDAR_START = Timestamp('2015-06-01', tz='UTC')
TEST_CALENDAR_STOP = Timestamp('2015-06-30', tz='UTC')

TEST_QUERY_START = Timestamp('2015-06-10', tz='UTC')
TEST_QUERY_STOP = Timestamp('2015-06-19', tz='UTC')

# One asset for each of the cases enumerated in load_raw_arrays_from_bcolz.
EQUITY_INFO = DataFrame(
    [
        # 1) The equity's trades start and end before query.
        {'start_date': '2015-06-01', 'end_date': '2015-06-05'},
        # 2) The equity's trades start and end after query.
        {'start_date': '2015-06-22', 'end_date': '2015-06-30'},
        # 3) The equity's data covers all dates in range.
        {'start_date': '2015-06-02', 'end_date': '2015-06-30'},
        # 4) The equity's trades start before the query start, but stop
        #    before the query end.
        {'start_date': '2015-06-01', 'end_date': '2015-06-15'},
        # 5) The equity's trades start and end during the query.
        {'start_date': '2015-06-12', 'end_date': '2015-06-18'},
        # 6) The equity's trades start during the query, but extend through
        #    the whole query.
        {'start_date': '2015-06-15', 'end_date': '2015-06-25'},
    ],
    index=arange(1, 7),
    columns=['start_date', 'end_date'],
).astype(datetime64)
EQUITY_INFO['symbol'] = [chr(ord('A') + n) for n in range(len(EQUITY_INFO))]
EQUITY_INFO['exchange'] = 'TEST'

TEST_QUERY_SIDS = EQUITY_INFO.index


# ADJUSTMENTS use the following scheme to indicate information about the value
# upon inspection.
#
# 1s place is the equity
#
# 0.1s place is the action type, with:
#
# splits, 1
# mergers, 2
# dividends, 3
#
# 0.001s is the date
SPLITS = DataFrame(
    [
        # Before query range, should be excluded.
        {'effective_date': str_to_seconds('2015-06-03'),
         'ratio': 1.103,
         'sid': 1},
        # First day of query range, should be excluded.
        {'effective_date': str_to_seconds('2015-06-10'),
         'ratio': 3.110,
         'sid': 3},
        # Third day of query range, should have last_row of 2
        {'effective_date': str_to_seconds('2015-06-12'),
         'ratio': 3.112,
         'sid': 3},
        # After query range, should be excluded.
        {'effective_date': str_to_seconds('2015-06-21'),
         'ratio': 6.121,
         'sid': 6},
        # Another action in query range, should have last_row of 1
        {'effective_date': str_to_seconds('2015-06-11'),
         'ratio': 3.111,
         'sid': 3},
        # Last day of range.  Should have last_row of 7
        {'effective_date': str_to_seconds('2015-06-19'),
         'ratio': 3.119,
         'sid': 3},
    ],
    columns=['effective_date', 'ratio', 'sid'],
)


MERGERS = DataFrame(
    [
        # Before query range, should be excluded.
        {'effective_date': str_to_seconds('2015-06-03'),
         'ratio': 1.203,
         'sid': 1},
        # First day of query range, should be excluded.
        {'effective_date': str_to_seconds('2015-06-10'),
         'ratio': 3.210,
         'sid': 3},
        # Third day of query range, should have last_row of 2
        {'effective_date': str_to_seconds('2015-06-12'),
         'ratio': 3.212,
         'sid': 3},
        # After query range, should be excluded.
        {'effective_date': str_to_seconds('2015-06-25'),
         'ratio': 6.225,
         'sid': 6},
        # Another action in query range, should have last_row of 2
        {'effective_date': str_to_seconds('2015-06-12'),
         'ratio': 4.212,
         'sid': 4},
        # Last day of range.  Should have last_row of 7
        {'effective_date': str_to_seconds('2015-06-19'),
         'ratio': 3.219,
         'sid': 3},
    ],
    columns=['effective_date', 'ratio', 'sid'],
)


DIVIDENDS = DataFrame(
    [
        # Before query range, should be excluded.
        {'declared_date': Timestamp('2015-05-01', tz='UTC').to_datetime64(),
         'ex_date': Timestamp('2015-06-01', tz='UTC').to_datetime64(),
         'record_date': Timestamp('2015-06-03', tz='UTC').to_datetime64(),
         'pay_date': Timestamp('2015-06-05', tz='UTC').to_datetime64(),
         'amount': 90.0,
         'sid': 1},
        # First day of query range, should be excluded.
        {'declared_date': Timestamp('2015-06-01', tz='UTC').to_datetime64(),
         'ex_date': Timestamp('2015-06-10', tz='UTC').to_datetime64(),
         'record_date': Timestamp('2015-06-15', tz='UTC').to_datetime64(),
         'pay_date': Timestamp('2015-06-17', tz='UTC').to_datetime64(),
         'amount': 80.0,
         'sid': 3},
        # Third day of query range, should have last_row of 2
        {'declared_date': Timestamp('2015-06-01', tz='UTC').to_datetime64(),
         'ex_date': Timestamp('2015-06-12', tz='UTC').to_datetime64(),
         'record_date': Timestamp('2015-06-15', tz='UTC').to_datetime64(),
         'pay_date': Timestamp('2015-06-17', tz='UTC').to_datetime64(),
         'amount': 70.0,
         'sid': 3},
        # After query range, should be excluded.
        {'declared_date': Timestamp('2015-06-01', tz='UTC').to_datetime64(),
         'ex_date': Timestamp('2015-06-25', tz='UTC').to_datetime64(),
         'record_date': Timestamp('2015-06-28', tz='UTC').to_datetime64(),
         'pay_date': Timestamp('2015-06-30', tz='UTC').to_datetime64(),
         'amount': 60.0,
         'sid': 6},
        # Another action in query range, should have last_row of 3
        {'declared_date': Timestamp('2015-06-01', tz='UTC').to_datetime64(),
         'ex_date': Timestamp('2015-06-15', tz='UTC').to_datetime64(),
         'record_date': Timestamp('2015-06-18', tz='UTC').to_datetime64(),
         'pay_date': Timestamp('2015-06-20', tz='UTC').to_datetime64(),
         'amount': 50.0,
         'sid': 3},
        # Last day of range.  Should have last_row of 7
        {'declared_date': Timestamp('2015-06-01', tz='UTC').to_datetime64(),
         'ex_date': Timestamp('2015-06-19', tz='UTC').to_datetime64(),
         'record_date': Timestamp('2015-06-22', tz='UTC').to_datetime64(),
         'pay_date': Timestamp('2015-06-30', tz='UTC').to_datetime64(),
         'amount': 40.0,
         'sid': 3},
    ],
    columns=['declared_date',
             'ex_date',
             'record_date',
             'pay_date',
             'amount',
             'sid'],
)


DIVIDENDS_EXPECTED = DataFrame(
    [
        # Before query range, should be excluded.
        {'effective_date': str_to_seconds('2015-06-01'),
         'ratio': 0.1,
         'sid': 1},
        # First day of query range, should be excluded.
        {'effective_date': str_to_seconds('2015-06-10'),
         'ratio': 0.20,
         'sid': 3},
        # Third day of query range, should have last_row of 2
        {'effective_date': str_to_seconds('2015-06-12'),
         'ratio': 0.30,
         'sid': 3},
        # After query range, should be excluded.
        {'effective_date': str_to_seconds('2015-06-25'),
         'ratio': 0.40,
         'sid': 6},
        # Another action in query range, should have last_row of 3
        {'effective_date': str_to_seconds('2015-06-15'),
         'ratio': 0.50,
         'sid': 3},
        # Last day of range.  Should have last_row of 7
        {'effective_date': str_to_seconds('2015-06-19'),
         'ratio': 0.60,
         'sid': 3},
    ],
    columns=['effective_date', 'ratio', 'sid'],
)


class USEquityPricingLoaderTestCase(WithAdjustmentReader,
                                    ZiplineTestCase):
    START_DATE = TEST_CALENDAR_START
    END_DATE = TEST_CALENDAR_STOP
    asset_ids = 1, 2, 3

    @classmethod
    def make_equity_info(cls):
        return EQUITY_INFO

    @classmethod
    def make_splits_data(cls):
        return SPLITS

    @classmethod
    def make_mergers_data(cls):
        return MERGERS

    @classmethod
    def make_dividends_data(cls):
        return DIVIDENDS

    @classmethod
    def make_adjustment_writer_equity_daily_bar_reader(cls):
        return MockDailyBarReader(
            dates=cls.calendar_days_between(cls.START_DATE, cls.END_DATE),
        )

    @classmethod
    def make_equity_daily_bar_data(cls, country_code, sids):
        return make_bar_data(
            EQUITY_INFO,
            cls.equity_daily_bar_days,
        )

    @classmethod
    def init_class_fixtures(cls):
        super(USEquityPricingLoaderTestCase, cls).init_class_fixtures()
        cls.sids = TEST_QUERY_SIDS
        cls.asset_info = EQUITY_INFO

    def test_input_sanity(self):
        # Ensure that the input data doesn't contain adjustments during periods
        # where the corresponding asset didn't exist.
        for table in SPLITS, MERGERS:
            for eff_date_secs, _, sid in table.itertuples(index=False):
                eff_date = Timestamp(eff_date_secs, unit='s')
                asset_start, asset_end = EQUITY_INFO.ix[
                    sid, ['start_date', 'end_date']
                ]
                self.assertGreaterEqual(eff_date, asset_start)
                self.assertLessEqual(eff_date, asset_end)

    @classmethod
    def calendar_days_between(cls, start_date, end_date, shift=0):
        slice_ = cls.equity_daily_bar_days.slice_indexer(start_date, end_date)
        start = slice_.start + shift
        stop = slice_.stop + shift
        if start < 0:
            raise KeyError(start_date, shift)

        return cls.equity_daily_bar_days[start:stop]

    def expected_adjustments(self,
                             start_date,
                             end_date,
                             tables,
                             adjustment_type):
        price_adjustments = {}
        volume_adjustments = {}

        should_include_price_adjustments = (
            adjustment_type == 'all' or adjustment_type == 'price'
        )
        should_include_volume_adjustments = (
            adjustment_type == 'all' or adjustment_type == 'volume'
        )

        query_days = self.calendar_days_between(start_date, end_date)
        start_loc = query_days.get_loc(start_date)

        for table in tables:
            for eff_date_secs, ratio, sid in table.itertuples(index=False):
                eff_date = Timestamp(eff_date_secs, unit='s', tz='UTC')

                # Ignore adjustments outside the query bounds.
                if not (start_date <= eff_date <= end_date):
                    continue

                eff_date_loc = query_days.get_loc(eff_date)
                delta = eff_date_loc - start_loc

                # Pricing adjustments should be applied on the date
                # corresponding to the effective date of the input data. They
                # should affect all rows **before** the effective date.
                if should_include_price_adjustments:
                    price_adjustments.setdefault(delta, []).append(
                        Float64Multiply(
                            first_row=0,
                            last_row=delta,
                            first_col=sid - 1,
                            last_col=sid - 1,
                            value=ratio,
                        )
                    )
                # Volume is *inversely* affected by *splits only*.
                if table is SPLITS and should_include_volume_adjustments:
                    volume_adjustments.setdefault(delta, []).append(
                        Float64Multiply(
                            first_row=0,
                            last_row=delta,
                            first_col=sid - 1,
                            last_col=sid - 1,
                            value=1.0 / ratio,
                        )
                    )

        output = {}
        if should_include_price_adjustments:
            output['price_adjustments'] = price_adjustments
        if should_include_volume_adjustments:
            output['volume_adjustments'] = volume_adjustments

        return output

    @parameterized([
        ([SPLITS, MERGERS, DIVIDENDS_EXPECTED], 'all'),
        ([SPLITS, MERGERS, DIVIDENDS_EXPECTED], 'price'),
        ([SPLITS, MERGERS, DIVIDENDS_EXPECTED], 'volume'),
        ([SPLITS, MERGERS, None], 'all'),
        ([SPLITS, MERGERS, None], 'price'),
    ])
    def test_load_adjustments(self, tables, adjustment_type):
        query_days = self.calendar_days_between(
            TEST_QUERY_START,
            TEST_QUERY_STOP,
        )

        adjustments = self.adjustment_reader.load_adjustments(
            query_days,
            self.sids,
            should_include_splits=tables[0] is not None,
            should_include_mergers=tables[1] is not None,
            should_include_dividends=tables[2] is not None,
            adjustment_type=adjustment_type,
        )
        expected_adjustments = self.expected_adjustments(
            TEST_QUERY_START,
            TEST_QUERY_STOP,
            [table for table in tables if table is not None],
            adjustment_type,
        )

        if adjustment_type == 'all' or adjustment_type == 'price':
            expected_price_adjustments = expected_adjustments['price']
            for key in expected_price_adjustments:
                price_adjustment = adjustments['price'][key]
                for j, adj in enumerate(price_adjustment):
                    expected = expected_price_adjustments[key][j]
                    self.assertEqual(adj.first_row, expected.first_row)
                    self.assertEqual(adj.last_row, expected.last_row)
                    self.assertEqual(adj.first_col, expected.first_col)
                    self.assertEqual(adj.last_col, expected.last_col)
                    assert_allclose(adj.value, expected.value)

        if adjustment_type == 'all' or adjustment_type == 'volume':
            expected_volume_adjustments = expected_adjustments['volume']
            for key in expected_volume_adjustments:
                volume_adjustment = adjustments['volume'][key]
                for j, adj in enumerate(volume_adjustment):
                    expected = expected_volume_adjustments[key][j]
                    self.assertEqual(adj.first_row, expected.first_row)
                    self.assertEqual(adj.last_row, expected.last_row)
                    self.assertEqual(adj.first_col, expected.first_col)
                    self.assertEqual(adj.last_col, expected.last_col)
                    assert_allclose(adj.value, expected.value)

    @parameterized([(True,), (False,)])
    def test_load_adjustments_to_df(self, convert_dts):
        reader = self.adjustment_reader
        adjustment_dfs = reader.unpack_db_to_component_dfs(
            convert_dates=convert_dts
        )

        name_and_raw = (
            ('splits', SPLITS),
            ('mergers', MERGERS),
            ('dividends', DIVIDENDS_EXPECTED)
        )

        def create_expected_table(df, name):
            expected_df = df.copy()

            if convert_dts:
                for colname in reader._datetime_int_cols[name]:
                    expected_df[colname] = expected_df[colname].astype(
                        'datetime64[s]'
                    )

            return expected_df

        def create_expected_div_table(df, name):
            expected_df = df.copy()

            if not convert_dts:
                for colname in reader._datetime_int_cols[name]:
                    expected_df[colname] = expected_df[colname].astype(
                        'datetime64[s]'
                    ).astype(int)

            return expected_df

        for action_name, raw_tbl in name_and_raw:

            exp = create_expected_table(raw_tbl, action_name)
            assert_frame_equal(
                adjustment_dfs[action_name],
                exp
            )

        # DIVIDENDS is in the opposite form from the rest of the dataframes, so
        # needs to be converted separately.
        div_name = 'dividend_payouts'
        assert_frame_equal(
            adjustment_dfs[div_name],
            create_expected_div_table(DIVIDENDS, div_name)
        )

    def test_read_no_adjustments(self):
        adjustment_reader = NullAdjustmentReader()
        columns = [USEquityPricing.close, USEquityPricing.volume]
        query_days = self.calendar_days_between(
            TEST_QUERY_START,
            TEST_QUERY_STOP
        )
        # Our expected results for each day are based on values from the
        # previous day.
        shifted_query_days = self.calendar_days_between(
            TEST_QUERY_START,
            TEST_QUERY_STOP,
            shift=-1,
        )

        adjustments = adjustment_reader.load_pricing_adjustments(
            [c.name for c in columns],
            query_days,
            self.sids,
        )
        self.assertEqual(adjustments, [{}, {}])

        pricing_loader = USEquityPricingLoader.without_fx(
            self.bcolz_equity_daily_bar_reader,
            adjustment_reader,
        )

        results = pricing_loader.load_adjusted_array(
            domain=US_EQUITIES,
            columns=columns,
            dates=query_days,
            sids=self.sids,
            mask=ones((len(query_days), len(self.sids)), dtype=bool),
        )
        closes, volumes = map(getitem(results), columns)

        expected_baseline_closes = expected_bar_values_2d(
            shifted_query_days,
            self.sids,
            self.asset_info,
            'close',
        )
        expected_baseline_volumes = expected_bar_values_2d(
            shifted_query_days,
            self.sids,
            self.asset_info,
            'volume',
        )

        # AdjustedArrays should yield the same data as the expected baseline.
        for windowlen in range(1, len(query_days) + 1):
            for offset, window in enumerate(closes.traverse(windowlen)):
                assert_array_equal(
                    expected_baseline_closes[offset:offset + windowlen],
                    window,
                )

            for offset, window in enumerate(volumes.traverse(windowlen)):
                assert_array_equal(
                    expected_baseline_volumes[offset:offset + windowlen],
                    window,
                )

        # Verify that we checked up to the longest possible window.
        with self.assertRaises(WindowLengthTooLong):
            closes.traverse(windowlen + 1)
        with self.assertRaises(WindowLengthTooLong):
            volumes.traverse(windowlen + 1)

    def apply_adjustments(self, dates, assets, baseline_values, adjustments):
        min_date, max_date = dates[[0, -1]]
        # HACK: Simulate the coercion to float64 we do in adjusted_array.  This
        # should be removed when AdjustedArray properly supports
        # non-floating-point types.
        orig_dtype = baseline_values.dtype
        values = baseline_values.astype(float64).copy()
        for eff_date_secs, ratio, sid in adjustments.itertuples(index=False):
            eff_date = seconds_to_timestamp(eff_date_secs)
            # Don't apply adjustments that aren't in the current date range.
            if eff_date not in dates:
                continue
            eff_date_loc = dates.get_loc(eff_date)
            asset_col = assets.get_loc(sid)
            # Apply ratio multiplicatively to the asset column on all rows less
            # than or equal adjustment effective date.
            values[:eff_date_loc + 1, asset_col] *= ratio
        return values.astype(orig_dtype)

    def test_read_with_adjustments(self):
        columns = [USEquityPricing.high, USEquityPricing.volume]
        query_days = self.calendar_days_between(
            TEST_QUERY_START,
            TEST_QUERY_STOP
        )
        # Our expected results for each day are based on values from the
        # previous day.
        shifted_query_days = self.calendar_days_between(
            TEST_QUERY_START,
            TEST_QUERY_STOP,
            shift=-1,
        )

        pricing_loader = USEquityPricingLoader.without_fx(
            self.bcolz_equity_daily_bar_reader,
            self.adjustment_reader,
        )

        results = pricing_loader.load_adjusted_array(
            domain=US_EQUITIES,
            columns=columns,
            dates=query_days,
            sids=Int64Index(arange(1, 7)),
            mask=ones((len(query_days), 6), dtype=bool),
        )
        highs, volumes = map(getitem(results), columns)

        expected_baseline_highs = expected_bar_values_2d(
            shifted_query_days,
            self.sids,
            self.asset_info,
            'high',
        )
        expected_baseline_volumes = expected_bar_values_2d(
            shifted_query_days,
            self.sids,
            self.asset_info,
            'volume',
        )

        # At each point in time, the AdjustedArrays should yield the baseline
        # with all adjustments up to that date applied.
        for windowlen in range(1, len(query_days) + 1):
            for offset, window in enumerate(highs.traverse(windowlen)):
                baseline = expected_baseline_highs[offset:offset + windowlen]
                baseline_dates = query_days[offset:offset + windowlen]
                expected_adjusted_highs = self.apply_adjustments(
                    baseline_dates,
                    self.sids,
                    baseline,
                    # Apply all adjustments.
                    concat([SPLITS, MERGERS, DIVIDENDS_EXPECTED],
                           ignore_index=True),
                )
                assert_allclose(expected_adjusted_highs, window)

            for offset, window in enumerate(volumes.traverse(windowlen)):
                baseline = expected_baseline_volumes[offset:offset + windowlen]
                baseline_dates = query_days[offset:offset + windowlen]
                # Apply only splits and invert the ratio.
                adjustments = SPLITS.copy()
                adjustments.ratio = 1 / adjustments.ratio

                expected_adjusted_volumes = self.apply_adjustments(
                    baseline_dates,
                    self.sids,
                    baseline,
                    adjustments,
                )
                # FIXME: Make AdjustedArray properly support integral types.
                assert_array_equal(
                    expected_adjusted_volumes,
                    window.astype(uint32),
                )

        # Verify that we checked up to the longest possible window.
        with self.assertRaises(WindowLengthTooLong):
            highs.traverse(windowlen + 1)
        with self.assertRaises(WindowLengthTooLong):
            volumes.traverse(windowlen + 1)



from six import iteritems

import numpy as np
import pandas as pd
from pandas import NaT

from trading_calendars import TradingCalendar

from zipline.data.bar_reader import OHLCV, NoDataOnDate, NoDataForSid
from zipline.data.session_bars import CurrencyAwareSessionBarReader
from zipline.utils.input_validation import expect_types, validate_keys
from zipline.utils.pandas_utils import check_indexes_all_same


class InMemoryDailyBarReader(CurrencyAwareSessionBarReader):
    """
    A SessionBarReader backed by a dictionary of in-memory DataFrames.

    Parameters
    ----------
    frames : dict[str -> pd.DataFrame]
        Dictionary from field name ("open", "high", "low", "close", or
        "volume") to DataFrame containing data for that field.
    calendar : str or trading_calendars.TradingCalendar
        Calendar (or name of calendar) to which data is aligned.
    currency_codes : pd.Series
        Map from sid -> listing currency for that sid.
    verify_indices : bool, optional
        Whether or not to verify that input data is correctly aligned to the
        given calendar. Default is True.
    """
    @expect_types(
        frames=dict,
        calendar=TradingCalendar,
        verify_indices=bool,
        currency_codes=pd.Series,
    )
    def __init__(self,
                 frames,
                 calendar,
                 currency_codes,
                 verify_indices=True):
        self._frames = frames
        self._values = {key: frame.values for key, frame in iteritems(frames)}
        self._calendar = calendar
        self._currency_codes = currency_codes

        validate_keys(frames, set(OHLCV), type(self).__name__)
        if verify_indices:
            verify_frames_aligned(list(frames.values()), calendar)

        self._sessions = frames['close'].index
        self._sids = frames['close'].columns

    @classmethod
    def from_panel(cls, panel, calendar, currency_codes):
        """Helper for construction from a pandas.Panel.
        """
        return cls(dict(panel.iteritems()), calendar, currency_codes)

    @property
    def last_available_dt(self):
        return self._calendar[-1]

    @property
    def trading_calendar(self):
        return self._calendar

    @property
    def sessions(self):
        return self._sessions

    def load_raw_arrays(self, columns, start_dt, end_dt, assets):
        if start_dt not in self._sessions:
            raise NoDataOnDate(start_dt)
        if end_dt not in self._sessions:
            raise NoDataOnDate(end_dt)

        asset_indexer = self._sids.get_indexer(assets)
        if -1 in asset_indexer:
            bad_assets = assets[asset_indexer == -1]
            raise NoDataForSid(bad_assets)

        date_indexer = self._sessions.slice_indexer(start_dt, end_dt)

        out = []
        for c in columns:
            out.append(self._values[c][date_indexer, asset_indexer])

        return out

    def get_value(self, sid, dt, field):
        """
        Parameters
        ----------
        sid : int
            The asset identifier.
        day : datetime64-like
            Midnight of the day for which data is requested.
        field : string
            The price field. e.g. ('open', 'high', 'low', 'close', 'volume')

        Returns
        -------
        float
            The spot price for colname of the given sid on the given day.
            Raises a NoDataOnDate exception if the given day and sid is before
            or after the date range of the equity.
            Returns -1 if the day is within the date range, but the price is
            0.
        """
        return self.frames[field].loc[dt, sid]

    def get_last_traded_dt(self, asset, dt):
        """
        Parameters
        ----------
        asset : zipline.asset.Asset
            The asset identifier.
        dt : datetime64-like
            Midnight of the day for which data is requested.

        Returns
        -------
        pd.Timestamp : The last know dt for the asset and dt;
                       NaT if no trade is found before the given dt.
        """
        try:
            return self.frames['close'].loc[:, asset.sid].last_valid_index()
        except IndexError:
            return NaT

    @property
    def first_trading_day(self):
        return self._sessions[0]

    def currency_codes(self, sids):
        codes = self._currency_codes
        return np.array([codes[sid] for sid in sids])


def verify_frames_aligned(frames, calendar):
    """
    Verify that DataFrames in ``frames`` have the same indexing scheme and are
    aligned to ``calendar``.

    Parameters
    ----------
    frames : list[pd.DataFrame]
    calendar : trading_calendars.TradingCalendar

    Raises
    ------
    ValueError
        If frames have different indexes/columns, or if frame indexes do not
        match a contiguous region of ``calendar``.
    """
    indexes = [f.index for f in frames]
    check_indexes_all_same(indexes, message="DataFrame indexes don't match:")

    columns = [f.columns for f in frames]
    check_indexes_all_same(columns, message="DataFrame columns don't match:")

    start, end = indexes[0][[0, -1]]
    cal_sessions = calendar.sessions_in_range(start, end)

    check_indexes_all_same(
        [indexes[0], cal_sessions],
        "DataFrame index doesn't match {} calendar:".format(calendar.name),
    )



from zipline.finance.metrics.core import _make_metrics_set_core
from zipline.testing.fixtures import ZiplineTestCase
from zipline.testing.predicates import (
    assert_equal,
    assert_is,
    assert_raises_str,
)
from zipline.utils.compat import mappingproxy


class MetricsSetCoreTestCase(ZiplineTestCase):
    def init_instance_fixtures(self):
        super(MetricsSetCoreTestCase, self).init_instance_fixtures()

        self.metrics_sets, self.register, self.unregister, self.load = (
            _make_metrics_set_core()
        )

        # make sure this starts empty
        assert_equal(self.metrics_sets, mappingproxy({}))

    def test_load_not_registered(self):
        msg = "no metrics set registered as 'ayy-lmao', options are: []"
        with assert_raises_str(ValueError, msg):
            self.load('ayy-lmao')

        # register in reverse order to test the sorting of the options
        self.register('c', set)
        self.register('b', set)
        self.register('a', set)

        msg = (
            "no metrics set registered as 'ayy-lmao', options are: "
            "['a', 'b', 'c']"
        )
        with assert_raises_str(ValueError, msg):
            self.load('ayy-lmao')

    def test_register_decorator(self):
        ayy_lmao_set = set()

        @self.register('ayy-lmao')
        def ayy_lmao():
            return ayy_lmao_set

        expected_metrics_sets = mappingproxy({'ayy-lmao': ayy_lmao})
        assert_equal(self.metrics_sets, expected_metrics_sets)
        assert_is(self.load('ayy-lmao'), ayy_lmao_set)

        msg = "metrics set 'ayy-lmao' is already registered"
        with assert_raises_str(ValueError, msg):
            @self.register('ayy-lmao')
            def other():  # pragma: no cover
                raise AssertionError('dead')

        # ensure that the failed registration didn't break the previously
        # registered set
        assert_equal(self.metrics_sets, expected_metrics_sets)
        assert_is(self.load('ayy-lmao'), ayy_lmao_set)

        self.unregister('ayy-lmao')
        assert_equal(self.metrics_sets, mappingproxy({}))

        msg = "no metrics set registered as 'ayy-lmao', options are: []"
        with assert_raises_str(ValueError, msg):
            self.load('ayy-lmao')

        msg = "metrics set 'ayy-lmao' was not already registered"
        with assert_raises_str(ValueError, msg):
            self.unregister('ayy-lmao')

    def test_register_non_decorator(self):
        ayy_lmao_set = set()

        def ayy_lmao():
            return ayy_lmao_set

        self.register('ayy-lmao', ayy_lmao)

        expected_metrics_sets = mappingproxy({'ayy-lmao': ayy_lmao})
        assert_equal(self.metrics_sets, expected_metrics_sets)
        assert_is(self.load('ayy-lmao'), ayy_lmao_set)

        def other():  # pragma: no cover
            raise AssertionError('dead')

        msg = "metrics set 'ayy-lmao' is already registered"
        with assert_raises_str(ValueError, msg):
            self.register('ayy-lmao', other)

        # ensure that the failed registration didn't break the previously
        # registered set
        assert_equal(self.metrics_sets, expected_metrics_sets)
        assert_is(self.load('ayy-lmao'), ayy_lmao_set)

        self.unregister('ayy-lmao')
        assert_equal(self.metrics_sets, mappingproxy({}))

        msg = "no metrics set registered as 'ayy-lmao', options are: []"
        with assert_raises_str(ValueError, msg):
            self.load('ayy-lmao')

        msg = "metrics set 'ayy-lmao' was not already registered"
        with assert_raises_str(ValueError, msg):
            self.unregister('ayy-lmao')



from collections import namedtuple

from .iface import PipelineHooks, PIPELINE_HOOKS_CONTEXT_MANAGERS

from interface import implements

from zipline.utils.compat import contextmanager, wraps


Call = namedtuple('Call', 'method_name args kwargs')


class ContextCall(namedtuple('ContextCall', 'state call')):

    @property
    def method_name(self):
        return self.call.method_name

    @property
    def args(self):
        return self.call.args

    @property
    def kwargs(self):
        return self.call.kwargs


def testing_hooks_method(method_name):
    """Factory function for making testing methods.
    """
    if method_name in PIPELINE_HOOKS_CONTEXT_MANAGERS:
        # Generate a method that enters the context of all sub-hooks.
        @wraps(getattr(PipelineHooks, method_name))
        @contextmanager
        def ctx(self, *args, **kwargs):
            call = Call(method_name, args, kwargs)
            self.trace.append(ContextCall('enter', call))
            yield
            self.trace.append(ContextCall('exit', call))
        return ctx

    else:
        # Generate a method that calls methods of all sub-hooks.
        @wraps(getattr(PipelineHooks, method_name))
        def method(self, *args, **kwargs):
            self.trace.append(Call(method_name, args, kwargs))
        return method


class TestingHooks(implements(PipelineHooks)):
    """A hooks implementation that keeps a trace of hook method calls.
    """
    def __init__(self):
        self.trace = []

    def clear(self):
        self.trace = []

    # Implement all interface methods by delegating to corresponding methods on
    # input hooks.
    locals().update({
        name: testing_hooks_method(name)
        # TODO: Expose this publicly on interface.
        for name in PipelineHooks._signatures
    })



from __future__ import division

import numpy as np
import pandas as pd
from trading_calendars import get_calendar

from zipline.data.bundles import ingest, load, bundles
from zipline.testing import test_resource_path
from zipline.testing.fixtures import ZiplineTestCase
from zipline.testing.predicates import assert_equal
from zipline.utils.functional import apply


class CSVDIRBundleTestCase(ZiplineTestCase):
    symbols = 'AAPL', 'IBM', 'KO', 'MSFT'
    asset_start = pd.Timestamp('2012-01-03', tz='utc')
    asset_end = pd.Timestamp('2014-12-31', tz='utc')
    bundle = bundles['csvdir']
    calendar = get_calendar(bundle.calendar_name)
    start_date = calendar.first_session
    end_date = calendar.last_session
    api_key = 'ayylmao'
    columns = 'open', 'high', 'low', 'close', 'volume'

    def _expected_data(self, asset_finder):
        sids = {
            symbol: asset_finder.lookup_symbol(
                symbol,
                self.asset_start,
            ).sid
            for symbol in self.symbols
        }

        def per_symbol(symbol):
            df = pd.read_csv(
                test_resource_path('csvdir_samples', 'csvdir',
                                   'daily', symbol + '.csv.gz'),
                parse_dates=['date'],
                index_col='date',
                usecols=[
                    'open',
                    'high',
                    'low',
                    'close',
                    'volume',
                    'date',
                    'dividend',
                    'split',
                ],
                na_values=['NA'],
            )
            df['sid'] = sids[symbol]
            return df

        all_ = pd.concat(map(per_symbol, self.symbols)).set_index(
            'sid',
            append=True,
        ).unstack()

        # fancy list comprehension with statements
        @list
        @apply
        def pricing():
            for column in self.columns:
                vs = all_[column].values
                if column == 'volume':
                    vs = np.nan_to_num(vs)
                yield vs

        adjustments = [[5572, 5576, 5595, 5634, 5639, 5659, 5698, 5699,
                        5701, 5702, 5722, 5760, 5764, 5774, 5821, 5822,
                        5829, 5845, 5884, 5885, 5888, 5908, 5947, 5948,
                        5951, 5972, 6011, 6020, 6026, 6073, 6080, 6096,
                        6135, 6136, 6139, 6157, 6160, 6198, 6199, 6207,
                        6223, 6263, 6271, 6277],
                       [5572, 5576, 5595, 5634, 5639, 5659, 5698, 5699,
                        5701, 5702, 5722, 5760, 5764, 5774, 5821, 5822,
                        5829, 5845, 5884, 5885, 5888, 5908, 5947, 5948,
                        5951, 5972, 6011, 6020, 6026, 6073, 6080, 6096,
                        6135, 6136, 6139, 6157, 6160, 6198, 6199, 6207,
                        6223, 6263, 6271, 6277],
                       [5572, 5576, 5595, 5634, 5639, 5659, 5698, 5699,
                        5701, 5702, 5722, 5760, 5764, 5774, 5821, 5822,
                        5829, 5845, 5884, 5885, 5888, 5908, 5947, 5948,
                        5951, 5972, 6011, 6020, 6026, 6073, 6080, 6096,
                        6135, 6136, 6139, 6157, 6160, 6198, 6199, 6207,
                        6223, 6263, 6271, 6277],
                       [5572, 5576, 5595, 5634, 5639, 5659, 5698, 5699,
                        5701, 5702, 5722, 5760, 5764, 5774, 5821, 5822,
                        5829, 5845, 5884, 5885, 5888, 5908, 5947, 5948,
                        5951, 5972, 6011, 6020, 6026, 6073, 6080, 6096,
                        6135, 6136, 6139, 6157, 6160, 6198, 6199, 6207,
                        6223, 6263, 6271, 6277],
                       [5701, 6157]]

        return pricing, adjustments

    def test_bundle(self):
        environ = {
            'CSVDIR': test_resource_path('csvdir_samples', 'csvdir')
        }

        ingest('csvdir', environ=environ)
        bundle = load('csvdir', environ=environ)
        sids = 0, 1, 2, 3
        assert_equal(set(bundle.asset_finder.sids), set(sids))

        for equity in bundle.asset_finder.retrieve_all(sids):
            assert_equal(equity.start_date, self.asset_start, msg=equity)
            assert_equal(equity.end_date, self.asset_end, msg=equity)

        sessions = self.calendar.all_sessions
        actual = bundle.equity_daily_bar_reader.load_raw_arrays(
            self.columns,
            sessions[sessions.get_loc(self.asset_start, 'bfill')],
            sessions[sessions.get_loc(self.asset_end, 'ffill')],
            sids,
        )

        expected_pricing, expected_adjustments = self._expected_data(
            bundle.asset_finder,
        )
        assert_equal(actual, expected_pricing, array_decimal=2)

        adjs_for_cols = bundle.adjustment_reader.load_pricing_adjustments(
            self.columns,
            sessions,
            pd.Index(sids),
        )
        assert_equal([sorted(adj.keys()) for adj in adjs_for_cols],
                     expected_adjustments)



#
# Copyright 2016 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from warnings import warn

import pandas as pd

from .assets import Asset
from .utils.enum import enum
from ._protocol import BarData, InnerPosition  # noqa


class MutableView(object):
    """A mutable view over an "immutable" object.

    Parameters
    ----------
    ob : any
        The object to take a view over.
    """
    # add slots so we don't accidentally add attributes to the view instead of
    # ``ob``
    __slots__ = ('_mutable_view_ob',)

    def __init__(self, ob):
        object.__setattr__(self, '_mutable_view_ob', ob)

    def __getattr__(self, attr):
        return getattr(self._mutable_view_ob, attr)

    def __setattr__(self, attr, value):
        vars(self._mutable_view_ob)[attr] = value

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, self._mutable_view_ob)


# Datasource type should completely determine the other fields of a
# message with its type.
DATASOURCE_TYPE = enum(
    'AS_TRADED_EQUITY',
    'MERGER',
    'SPLIT',
    'DIVIDEND',
    'TRADE',
    'TRANSACTION',
    'ORDER',
    'EMPTY',
    'DONE',
    'CUSTOM',
    'BENCHMARK',
    'COMMISSION',
    'CLOSE_POSITION'
)

# Expected fields/index values for a dividend Series.
DIVIDEND_FIELDS = [
    'declared_date',
    'ex_date',
    'gross_amount',
    'net_amount',
    'pay_date',
    'payment_sid',
    'ratio',
    'sid',
]
# Expected fields/index values for a dividend payment Series.
DIVIDEND_PAYMENT_FIELDS = [
    'id',
    'payment_sid',
    'cash_amount',
    'share_count',
]


class Event(object):

    def __init__(self, initial_values=None):
        if initial_values:
            self.__dict__.update(initial_values)

    def keys(self):
        return self.__dict__.keys()

    def __eq__(self, other):
        return hasattr(other, '__dict__') and self.__dict__ == other.__dict__

    def __contains__(self, name):
        return name in self.__dict__

    def __repr__(self):
        return "Event({0})".format(self.__dict__)

    def to_series(self, index=None):
        return pd.Series(self.__dict__, index=index)


def _deprecated_getitem_method(name, attrs):
    """Create a deprecated ``__getitem__`` method that tells users to use
    getattr instead.

    Parameters
    ----------
    name : str
        The name of the object in the warning message.
    attrs : iterable[str]
        The set of allowed attributes.

    Returns
    -------
    __getitem__ : callable[any, str]
        The ``__getitem__`` method to put in the class dict.
    """
    attrs = frozenset(attrs)
    msg = (
        "'{name}[{attr!r}]' is deprecated, please use"
        " '{name}.{attr}' instead"
    )

    def __getitem__(self, key):
        """``__getitem__`` is deprecated, please use attribute access instead.
        """
        warn(msg.format(name=name, attr=key), DeprecationWarning, stacklevel=2)
        if key in attrs:
            return getattr(self, key)
        raise KeyError(key)

    return __getitem__


class Order(Event):
    # If you are adding new attributes, don't update this set. This method
    # is deprecated to normal attribute access so we don't want to encourage
    # new usages.
    __getitem__ = _deprecated_getitem_method(
        'order', {
            'dt',
            'sid',
            'amount',
            'stop',
            'limit',
            'id',
            'filled',
            'commission',
            'stop_reached',
            'limit_reached',
            'created',
        },
    )


class Portfolio(object):
    """Object providing read-only access to current portfolio state.

    Parameters
    ----------
    start_date : pd.Timestamp
        The start date for the period being recorded.
    capital_base : float
        The starting value for the portfolio. This will be used as the starting
        cash, current cash, and portfolio value.

    Attributes
    ----------
    positions : zipline.protocol.Positions
        Dict-like object containing information about currently-held positions.
    cash : float
        Amount of cash currently held in portfolio.
    portfolio_value : float
        Current liquidation value of the portfolio's holdings.
        This is equal to ``cash + sum(shares * price)``
    starting_cash : float
        Amount of cash in the portfolio at the start of the backtest.
    """

    def __init__(self, start_date=None, capital_base=0.0):
        self_ = MutableView(self)
        self_.cash_flow = 0.0
        self_.starting_cash = capital_base
        self_.portfolio_value = capital_base
        self_.pnl = 0.0
        self_.returns = 0.0
        self_.cash = capital_base
        self_.positions = Positions()
        self_.start_date = start_date
        self_.positions_value = 0.0
        self_.positions_exposure = 0.0

    @property
    def capital_used(self):
        return self.cash_flow

    def __setattr__(self, attr, value):
        raise AttributeError('cannot mutate Portfolio objects')

    def __repr__(self):
        return "Portfolio({0})".format(self.__dict__)

    # If you are adding new attributes, don't update this set. This method
    # is deprecated to normal attribute access so we don't want to encourage
    # new usages.
    __getitem__ = _deprecated_getitem_method(
        'portfolio', {
            'capital_used',
            'starting_cash',
            'portfolio_value',
            'pnl',
            'returns',
            'cash',
            'positions',
            'start_date',
            'positions_value',
        },
    )

    @property
    def current_portfolio_weights(self):
        """
        Compute each asset's weight in the portfolio by calculating its held
        value divided by the total value of all positions.

        Each equity's value is its price times the number of shares held. Each
        futures contract's value is its unit price times number of shares held
        times the multiplier.
        """
        position_values = pd.Series({
            asset: (
                    position.last_sale_price *
                    position.amount *
                    asset.price_multiplier
            )
            for asset, position in self.positions.items()
        })
        return position_values / self.portfolio_value


class Account(object):
    """
    The account object tracks information about the trading account. The
    values are updated as the algorithm runs and its keys remain unchanged.
    If connected to a broker, one can update these values with the trading
    account values as reported by the broker.
    """

    def __init__(self):
        self_ = MutableView(self)
        self_.settled_cash = 0.0
        self_.accrued_interest = 0.0
        self_.buying_power = float('inf')
        self_.equity_with_loan = 0.0
        self_.total_positions_value = 0.0
        self_.total_positions_exposure = 0.0
        self_.regt_equity = 0.0
        self_.regt_margin = float('inf')
        self_.initial_margin_requirement = 0.0
        self_.maintenance_margin_requirement = 0.0
        self_.available_funds = 0.0
        self_.excess_liquidity = 0.0
        self_.cushion = 0.0
        self_.day_trades_remaining = float('inf')
        self_.leverage = 0.0
        self_.net_leverage = 0.0
        self_.net_liquidation = 0.0

    def __setattr__(self, attr, value):
        raise AttributeError('cannot mutate Account objects')

    def __repr__(self):
        return "Account({0})".format(self.__dict__)

    # If you are adding new attributes, don't update this set. This method
    # is deprecated to normal attribute access so we don't want to encourage
    # new usages.
    __getitem__ = _deprecated_getitem_method(
        'account', {
            'settled_cash',
            'accrued_interest',
            'buying_power',
            'equity_with_loan',
            'total_positions_value',
            'total_positions_exposure',
            'regt_equity',
            'regt_margin',
            'initial_margin_requirement',
            'maintenance_margin_requirement',
            'available_funds',
            'excess_liquidity',
            'cushion',
            'day_trades_remaining',
            'leverage',
            'net_leverage',
            'net_liquidation',
        },
    )


class Position(object):
    """
    A position held by an algorithm.

    Attributes
    ----------
    asset : zipline.assets.Asset
        The held asset.
    amount : int
        Number of shares held. Short positions are represented with negative
        values.
    cost_basis : float
        Average price at which currently-held shares were acquired.
    last_sale_price : float
        Most recent price for the position.
    last_sale_date : pd.Timestamp
        Datetime at which ``last_sale_price`` was last updated.
    """
    __slots__ = ('_underlying_position',)

    def __init__(self, underlying_position):
        object.__setattr__(self, '_underlying_position', underlying_position)

    def __getattr__(self, attr):
        return getattr(self._underlying_position, attr)

    def __setattr__(self, attr, value):
        raise AttributeError('cannot mutate Position objects')

    @property
    def sid(self):
        # for backwards compatibility
        return self.asset

    def __repr__(self):
        return 'Position(%r)' % {
            k: getattr(self, k)
            for k in (
                'asset',
                'amount',
                'cost_basis',
                'last_sale_price',
                'last_sale_date',
            )
        }

    # If you are adding new attributes, don't update this set. This method
    # is deprecated to normal attribute access so we don't want to encourage
    # new usages.
    __getitem__ = _deprecated_getitem_method(
        'position', {
            'sid',
            'amount',
            'cost_basis',
            'last_sale_price',
            'last_sale_date',
        },
    )


# Copied from Position and renamed.  This is used to handle cases where a user
# does something like `context.portfolio.positions[100]` instead of
# `context.portfolio.positions[sid(100)]`.
class _DeprecatedSidLookupPosition(object):
    def __init__(self, sid):
        self.sid = sid
        self.amount = 0
        self.cost_basis = 0.0  # per share
        self.last_sale_price = 0.0
        self.last_sale_date = None

    def __repr__(self):
        return "_DeprecatedSidLookupPosition({0})".format(self.__dict__)

    # If you are adding new attributes, don't update this set. This method
    # is deprecated to normal attribute access so we don't want to encourage
    # new usages.
    __getitem__ = _deprecated_getitem_method(
        'position', {
            'sid',
            'amount',
            'cost_basis',
            'last_sale_price',
            'last_sale_date',
        },
    )


class Positions(dict):
    """A dict-like object containing the algorithm's current positions.
    """

    def __missing__(self, key):
        if isinstance(key, Asset):
            return Position(InnerPosition(key))
        elif isinstance(key, int):
            warn("Referencing positions by integer is deprecated."
                 " Use an asset instead.")
        else:
            warn("Position lookup expected a value of type Asset but got {0}"
                 " instead.".format(type(key).__name__))

        return _DeprecatedSidLookupPosition(key)



#
# Copyright 2014 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest import TestCase

from zipline.utils.argcheck import (
    verify_callable_argspec,
    Argument,
    NoStarargs,
    UnexpectedStarargs,
    NoKwargs,
    UnexpectedKwargs,
    NotCallable,
    NotEnoughArguments,
    TooManyArguments,
    MismatchedArguments,
)


class TestArgCheck(TestCase):
    def test_not_callable(self):
        """
        Check the results of a non-callable object.
        """
        not_callable = 'a'

        with self.assertRaises(NotCallable):
            verify_callable_argspec(not_callable)

    def test_no_starargs(self):
        """
        Tests when a function does not have *args and it was expected.
        """
        def f(a):
            pass

        with self.assertRaises(NoStarargs):
            verify_callable_argspec(f, expect_starargs=True)

    def test_starargs(self):
        """
        Tests when a function has *args and it was expected.
        """
        def f(*args):
            pass

        verify_callable_argspec(f, expect_starargs=True)

    def test_unexcpected_starargs(self):
        """
        Tests a function that unexpectedly accepts *args.
        """
        def f(*args):
            pass

        with self.assertRaises(UnexpectedStarargs):
            verify_callable_argspec(f, expect_starargs=False)

    def test_ignore_starargs(self):
        """
        Tests checking a function ignoring the presence of *args.
        """
        def f(*args):
            pass

        def g():
            pass

        verify_callable_argspec(f, expect_starargs=Argument.ignore)
        verify_callable_argspec(g, expect_starargs=Argument.ignore)

    def test_no_kwargs(self):
        """
        Tests when a function does not have **kwargs and it was expected.
        """
        def f():
            pass

        with self.assertRaises(NoKwargs):
            verify_callable_argspec(f, expect_kwargs=True)

    def test_kwargs(self):
        """
        Tests when a function has **kwargs and it was expected.
        """
        def f(**kwargs):
            pass

        verify_callable_argspec(f, expect_kwargs=True)

    def test_unexpected_kwargs(self):
        """
        Tests a function that unexpectedly accepts **kwargs.
        """
        def f(**kwargs):
            pass

        with self.assertRaises(UnexpectedKwargs):
            verify_callable_argspec(f, expect_kwargs=False)

    def test_ignore_kwargs(self):
        """
        Tests checking a function ignoring the presence of **kwargs.
        """
        def f(**kwargs):
            pass

        def g():
            pass

        verify_callable_argspec(f, expect_kwargs=Argument.ignore)
        verify_callable_argspec(g, expect_kwargs=Argument.ignore)

    def test_arg_subset(self):
        """
        Tests when the args are a subset of the expectations.
        """
        def f(a, b):
            pass

        with self.assertRaises(NotEnoughArguments):
            verify_callable_argspec(
                f, [Argument('a'), Argument('b'), Argument('c')]
            )

    def test_arg_superset(self):
        def f(a, b, c):
            pass

        with self.assertRaises(TooManyArguments):
            verify_callable_argspec(f, [Argument('a'), Argument('b')])

    def test_no_default(self):
        """
        Tests when an argument expects a default and it is not present.
        """
        def f(a):
            pass

        with self.assertRaises(MismatchedArguments):
            verify_callable_argspec(f, [Argument('a', 1)])

    def test_default(self):
        """
        Tests when an argument expects a default and it is present.
        """
        def f(a=1):
            pass

        verify_callable_argspec(f, [Argument('a', 1)])

    def test_ignore_default(self):
        """
        Tests that ignoring defaults works as intended.
        """
        def f(a=1):
            pass

        verify_callable_argspec(f, [Argument('a')])

    def test_mismatched_args(self):
        def f(a, b):
            pass

        with self.assertRaises(MismatchedArguments):
            verify_callable_argspec(f, [Argument('c'), Argument('d')])

    def test_ignore_args(self):
        """
        Tests the ignore argument list feature.
        """
        def f(a):
            pass

        def g():
            pass

        h = 'not_callable'

        verify_callable_argspec(f)
        verify_callable_argspec(g)
        with self.assertRaises(NotCallable):
            verify_callable_argspec(h)

    def test_out_of_order(self):
        """
        Tests the case where arguments are not in the correct order.
        """
        def f(a, b):
            pass

        with self.assertRaises(MismatchedArguments):
            verify_callable_argspec(f, [Argument('b'), Argument('a')])

    def test_wrong_default(self):
        """
        Tests the case where a default is expected, but the default provided
        does not match the one expected.
        """
        def f(a=1):
            pass

        with self.assertRaises(MismatchedArguments):
            verify_callable_argspec(f, [Argument('a', 2)])

    def test_any_default(self):
        """
        Tests the any_default option.
        """
        def f(a=1):
            pass

        def g(a=2):
            pass

        def h(a):
            pass

        expected_args = [Argument('a', Argument.any_default)]
        verify_callable_argspec(f, expected_args)
        verify_callable_argspec(g, expected_args)
        with self.assertRaises(MismatchedArguments):
            verify_callable_argspec(h, expected_args)

    def test_ignore_name(self):
        """
        Tests ignoring a param name.
        """
        def f(a):
            pass

        def g(b):
            pass

        def h(c=1):
            pass

        expected_args = [Argument(Argument.ignore, Argument.no_default)]
        verify_callable_argspec(f, expected_args)
        verify_callable_argspec(f, expected_args)
        with self.assertRaises(MismatchedArguments):
            verify_callable_argspec(h, expected_args)

    def test_bound_method(self):
        class C(object):
            def f(self, a, b):
                pass

        method = C().f

        verify_callable_argspec(method, [Argument('a'), Argument('b')])
        with self.assertRaises(NotEnoughArguments):
            # Assert that we don't count self.
            verify_callable_argspec(
                method,
                [Argument('self'), Argument('a'), Argument('b')],
            )



from datetime import time
from unittest import TestCase
import pandas as pd
from trading_calendars import get_calendar
from trading_calendars.utils.pandas_utils import days_at_time

from zipline.gens.sim_engine import (
    MinuteSimulationClock,
    SESSION_START,
    BEFORE_TRADING_START_BAR,
    BAR,
    SESSION_END
)


class TestClock(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.nyse_calendar = get_calendar("NYSE")

        # july 15 is friday, so there are 3 sessions in this range (15, 18, 19)
        cls.sessions = cls.nyse_calendar.sessions_in_range(
            pd.Timestamp("2016-07-15"),
            pd.Timestamp("2016-07-19")
        )

        trading_o_and_c = cls.nyse_calendar.schedule.ix[cls.sessions]
        cls.opens = trading_o_and_c['market_open']
        cls.closes = trading_o_and_c['market_close']

    def test_bts_before_session(self):
        clock = MinuteSimulationClock(
            self.sessions,
            self.opens,
            self.closes,
            days_at_time(self.sessions, time(6, 17), "US/Eastern"),
            False
        )

        all_events = list(clock)

        def _check_session_bts_first(session_label, events, bts_dt):
            minutes = self.nyse_calendar.minutes_for_session(session_label)

            self.assertEqual(393, len(events))

            self.assertEqual(events[0], (session_label, SESSION_START))
            self.assertEqual(events[1], (bts_dt, BEFORE_TRADING_START_BAR))
            for i in range(2, 392):
                self.assertEqual(events[i], (minutes[i - 2], BAR))
            self.assertEqual(events[392], (minutes[-1], SESSION_END))

        _check_session_bts_first(
            self.sessions[0],
            all_events[0:393],
            pd.Timestamp("2016-07-15 6:17", tz='US/Eastern')
        )

        _check_session_bts_first(
            self.sessions[1],
            all_events[393:786],
            pd.Timestamp("2016-07-18 6:17", tz='US/Eastern')
        )

        _check_session_bts_first(
            self.sessions[2],
            all_events[786:],
            pd.Timestamp("2016-07-19 6:17", tz='US/Eastern')
        )

    def test_bts_during_session(self):
        self.verify_bts_during_session(
            time(11, 45), [
                pd.Timestamp("2016-07-15 11:45", tz='US/Eastern'),
                pd.Timestamp("2016-07-18 11:45", tz='US/Eastern'),
                pd.Timestamp("2016-07-19 11:45", tz='US/Eastern')
            ],
            135
        )

    def test_bts_on_first_minute(self):
        self.verify_bts_during_session(
            time(9, 30), [
                pd.Timestamp("2016-07-15 9:30", tz='US/Eastern'),
                pd.Timestamp("2016-07-18 9:30", tz='US/Eastern'),
                pd.Timestamp("2016-07-19 9:30", tz='US/Eastern')
            ],
            1
        )

    def test_bts_on_last_minute(self):
        self.verify_bts_during_session(
            time(16, 00), [
                pd.Timestamp("2016-07-15 16:00", tz='US/Eastern'),
                pd.Timestamp("2016-07-18 16:00", tz='US/Eastern'),
                pd.Timestamp("2016-07-19 16:00", tz='US/Eastern')
            ],
            390
        )

    def verify_bts_during_session(self, bts_time, bts_session_times, bts_idx):
        def _check_session_bts_during(session_label, events, bts_dt):
            minutes = self.nyse_calendar.minutes_for_session(session_label)

            self.assertEqual(393, len(events))

            self.assertEqual(events[0], (session_label, SESSION_START))

            for i in range(1, bts_idx):
                self.assertEqual(events[i], (minutes[i - 1], BAR))

            self.assertEqual(
                events[bts_idx],
                (bts_dt, BEFORE_TRADING_START_BAR)
            )

            for i in range(bts_idx + 1, 391):
                self.assertEqual(events[i], (minutes[i - 2], BAR))

            self.assertEqual(events[392], (minutes[-1], SESSION_END))

        clock = MinuteSimulationClock(
            self.sessions,
            self.opens,
            self.closes,
            days_at_time(self.sessions, bts_time, "US/Eastern"),
            False
        )

        all_events = list(clock)

        _check_session_bts_during(
            self.sessions[0],
            all_events[0:393],
            bts_session_times[0]
        )

        _check_session_bts_during(
            self.sessions[1],
            all_events[393:786],
            bts_session_times[1]
        )

        _check_session_bts_during(
            self.sessions[2],
            all_events[786:],
            bts_session_times[2]
        )

    def test_bts_after_session(self):
        clock = MinuteSimulationClock(
            self.sessions,
            self.opens,
            self.closes,
            days_at_time(self.sessions, time(19, 5), "US/Eastern"),
            False
        )

        all_events = list(clock)

        # since 19:05 Eastern is after the NYSE is closed, we don't emit
        # BEFORE_TRADING_START.  therefore, each day has SESSION_START,
        # 390 BARs, and then SESSION_END

        def _check_session_bts_after(session_label, events):
            minutes = self.nyse_calendar.minutes_for_session(session_label)

            self.assertEqual(392, len(events))
            self.assertEqual(events[0], (session_label, SESSION_START))

            for i in range(1, 391):
                self.assertEqual(events[i], (minutes[i - 1], BAR))

            self.assertEqual(events[-1], (minutes[389], SESSION_END))

        for i in range(0, 2):
            _check_session_bts_after(
                self.sessions[i],
                all_events[(i * 392): ((i + 1) * 392)]
            )



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta, abstractmethod
from six import with_metaclass
from zipline.extensions import extensible
from zipline.finance.cancel_policy import NeverCancel


@extensible
class Blotter(with_metaclass(ABCMeta)):

    def __init__(self, cancel_policy=None):
        self.cancel_policy = cancel_policy if cancel_policy else NeverCancel()
        self.current_dt = None

    def set_date(self, dt):
        self.current_dt = dt

    @abstractmethod
    def order(self, asset, amount, style, order_id=None):
        """Place an order.

        Parameters
        ----------
        asset : zipline.assets.Asset
            The asset that this order is for.
        amount : int
            The amount of shares to order. If ``amount`` is positive, this is
            the number of shares to buy or cover. If ``amount`` is negative,
            this is the number of shares to sell or short.
        style : zipline.finance.execution.ExecutionStyle
            The execution style for the order.
        order_id : str, optional
            The unique identifier for this order.

        Returns
        -------
        order_id : str or None
            The unique identifier for this order, or None if no order was
            placed.

        Notes
        -----
        amount > 0 :: Buy/Cover
        amount < 0 :: Sell/Short
        Market order:    order(asset, amount)
        Limit order:     order(asset, amount, style=LimitOrder(limit_price))
        Stop order:      order(asset, amount, style=StopOrder(stop_price))
        StopLimit order: order(asset, amount, style=StopLimitOrder(limit_price,
                               stop_price))
        """

        raise NotImplementedError('order')

    def batch_order(self, order_arg_lists):
        """Place a batch of orders.

        Parameters
        ----------
        order_arg_lists : iterable[tuple]
            Tuples of args that `order` expects.

        Returns
        -------
        order_ids : list[str or None]
            The unique identifier (or None) for each of the orders placed
            (or not placed).

        Notes
        -----
        This is required for `Blotter` subclasses to be able to place a batch
        of orders, instead of being passed the order requests one at a time.
        """

        return [self.order(*order_args) for order_args in order_arg_lists]

    @abstractmethod
    def cancel(self, order_id, relay_status=True):
        """Cancel a single order

        Parameters
        ----------
        order_id : int
            The id of the order

        relay_status : bool
            Whether or not to record the status of the order
        """
        raise NotImplementedError('cancel')

    @abstractmethod
    def cancel_all_orders_for_asset(self, asset, warn=False,
                                    relay_status=True):
        """
        Cancel all open orders for a given asset.
        """

        raise NotImplementedError('cancel_all_orders_for_asset')

    @abstractmethod
    def execute_cancel_policy(self, event):
        raise NotImplementedError('execute_cancel_policy')

    @abstractmethod
    def reject(self, order_id, reason=''):
        """
        Mark the given order as 'rejected', which is functionally similar to
        cancelled. The distinction is that rejections are involuntary (and
        usually include a message from a broker indicating why the order was
        rejected) while cancels are typically user-driven.
        """

        raise NotImplementedError('reject')

    @abstractmethod
    def hold(self, order_id, reason=''):
        """
        Mark the order with order_id as 'held'. Held is functionally similar
        to 'open'. When a fill (full or partial) arrives, the status
        will automatically change back to open/filled as necessary.
        """

        raise NotImplementedError('hold')

    @abstractmethod
    def process_splits(self, splits):
        """
        Processes a list of splits by modifying any open orders as needed.

        Parameters
        ----------
        splits: list
            A list of splits.  Each split is a tuple of (asset, ratio).

        Returns
        -------
        None
        """

        raise NotImplementedError('process_splits')

    @abstractmethod
    def get_transactions(self, bar_data):
        """
        Creates a list of transactions based on the current open orders,
        slippage model, and commission model.

        Parameters
        ----------
        bar_data: zipline._protocol.BarData

        Notes
        -----
        This method book-keeps the blotter's open_orders dictionary, so that
         it is accurate by the time we're done processing open orders.

        Returns
        -------
        transactions_list: List
            transactions_list: list of transactions resulting from the current
            open orders.  If there were no open orders, an empty list is
            returned.

        commissions_list: List
            commissions_list: list of commissions resulting from filling the
            open orders.  A commission is an object with "asset" and "cost"
            parameters.

        closed_orders: List
            closed_orders: list of all the orders that have filled.
        """

        raise NotImplementedError('get_transactions')

    @abstractmethod
    def prune_orders(self, closed_orders):
        """
        Removes all given orders from the blotter's open_orders list.

        Parameters
        ----------
        closed_orders: iterable of orders that are closed.

        Returns
        -------
        None
        """

        raise NotImplementedError('prune_orders')



"""
Tests for the blaze interface to the pipeline api.
"""
from __future__ import division

from collections import OrderedDict
from datetime import timedelta, time
from functools import partial
from itertools import product, chain
from unittest import skipIf
import warnings

import blaze as bz
from datashape import dshape, var, Record
from nose_parameterized import parameterized
import numpy as np
from numpy.testing.utils import assert_array_almost_equal
from odo import odo
import pandas as pd
import pytz
from toolz import keymap, valmap, concatv
from toolz.curried import operator as op

from zipline.assets.synthetic import make_simple_equity_info
from zipline.errors import UnsupportedPipelineOutput
from zipline.pipeline import Pipeline, CustomFactor
from zipline.pipeline.data import DataSet, BoundColumn, Column
from zipline.pipeline.domain import EquitySessionDomain
from zipline.pipeline.engine import SimplePipelineEngine
from zipline.pipeline.loaders.blaze import (
    from_blaze,
    BlazeLoader,
    NoMetaDataWarning,
)
from zipline.pipeline.loaders.blaze.core import (
    ExprData,
    NonPipelineField,
)
from zipline.testing import (
    ZiplineTestCase,
    parameter_space,
    tmp_asset_finder,
)
from zipline.testing.fixtures import WithAssetFinder
from zipline.testing.predicates import (
    assert_equal,
    assert_frame_equal,
    assert_isidentical,
)
from zipline.utils.numpy_utils import float64_dtype, int64_dtype
from zipline.utils.pandas_utils import empty_dataframe, new_pandas, \
    skip_pipeline_new_pandas

nameof = op.attrgetter('name')
dtypeof = op.attrgetter('dtype')
asset_infos = (
    (make_simple_equity_info(
        tuple(map(ord, 'ABC')),
        pd.Timestamp(0),
        pd.Timestamp('2015'),
    ),),
    (make_simple_equity_info(
        tuple(map(ord, 'ABCD')),
        pd.Timestamp(0),
        pd.Timestamp('2015'),
    ),),
)
simple_asset_info = asset_infos[0][0]


def with_extra_sid():
    return parameterized.expand(asset_infos)


def with_ignore_sid():
    return parameterized.expand(
        product(chain.from_iterable(asset_infos), [True, False])
    )


class BlazeToPipelineTestCase(WithAssetFinder, ZiplineTestCase):
    START_DATE = pd.Timestamp(0)
    END_DATE = pd.Timestamp('2015')

    @classmethod
    def init_class_fixtures(cls):
        super(BlazeToPipelineTestCase, cls).init_class_fixtures()
        cls.dates = dates = pd.date_range('2014-01-01', '2014-01-03')
        cls.asof_dates = asof_dates = dates - pd.Timedelta(days=1)
        cls.timestamps = timestamps = dates - pd.Timedelta(hours=1)

        cls.df = df = pd.DataFrame({
            'sid': cls.ASSET_FINDER_EQUITY_SIDS * 3,
            'value': (0., 1., 2., 1., 2., 3., 2., 3., 4.),
            'int_value': (0, 1, 2, 1, 2, 3, 2, 3, 4),
            'asof_date': asof_dates.repeat(3),
            'timestamp': timestamps.repeat(3),
        })
        cls.dshape = dshape("""
        var * {
            sid: ?int64,
            value: ?float64,
            int_value: ?int64,
            asof_date: datetime,
            timestamp: datetime
        }
        """)
        cls.macro_df = df[df.sid == 65].drop('sid', axis=1)
        dshape_ = OrderedDict(cls.dshape.measure.fields)
        del dshape_['sid']
        cls.macro_dshape = var * Record(dshape_)

        cls.garbage_loader = BlazeLoader()
        cls.missing_values = {'int_value': 0}

        cls.value_dshape = dshape("""var * {
            sid: ?int64,
            value: float64,
            asof_date: datetime,
            timestamp: datetime,
        }""")

    def create_domain(self,
                      sessions,
                      data_query_time=time(0, 0, tzinfo=pytz.utc),
                      data_query_date_offset=0):
        if sessions.tz is None:
            sessions = sessions.tz_localize('UTC')

        return EquitySessionDomain(
            sessions,
            country_code=self.ASSET_FINDER_COUNTRY_CODE,
            data_query_time=data_query_time,
            data_query_date_offset=data_query_date_offset,
        )

    def test_tabular(self):
        name = 'expr'
        expr = bz.data(self.df, name=name, dshape=self.dshape)
        ds = from_blaze(
            expr,
            loader=self.garbage_loader,
            no_deltas_rule='ignore',
            no_checkpoints_rule='ignore',
            missing_values=self.missing_values,
        )
        self.assertEqual(ds.__name__, name)
        self.assertTrue(issubclass(ds, DataSet))

        self.assertIs(ds.value.dtype, float64_dtype)
        self.assertIs(ds.int_value.dtype, int64_dtype)

        self.assertTrue(np.isnan(ds.value.missing_value))
        self.assertEqual(ds.int_value.missing_value, 0)

        # test memoization
        self.assertIs(
            from_blaze(
                expr,
                loader=self.garbage_loader,
                no_deltas_rule='ignore',
                no_checkpoints_rule='ignore',
                missing_values=self.missing_values,
            ),
            ds,
        )

    def test_column(self):
        exprname = 'expr'
        expr = bz.data(self.df, name=exprname, dshape=self.dshape)
        value = from_blaze(
            expr.value,
            loader=self.garbage_loader,
            no_deltas_rule='ignore',
            no_checkpoints_rule='ignore',
            missing_values=self.missing_values,
        )
        self.assertEqual(value.name, 'value')
        self.assertIsInstance(value, BoundColumn)
        self.assertIs(value.dtype, float64_dtype)

        # test memoization
        self.assertIs(
            from_blaze(
                expr.value,
                loader=self.garbage_loader,
                no_deltas_rule='ignore',
                no_checkpoints_rule='ignore',
                missing_values=self.missing_values,
            ),
            value,
        )
        self.assertIs(
            from_blaze(
                expr,
                loader=self.garbage_loader,
                no_deltas_rule='ignore',
                no_checkpoints_rule='ignore',
                missing_values=self.missing_values,
            ).value,
            value,
        )

        # test the walk back up the tree
        self.assertIs(
            from_blaze(
                expr,
                loader=self.garbage_loader,
                no_deltas_rule='ignore',
                no_checkpoints_rule='ignore',
                missing_values=self.missing_values,
            ),
            value.dataset,
        )
        self.assertEqual(value.dataset.__name__, exprname)

    def test_missing_asof(self):
        expr = bz.data(
            self.df.loc[:, ['sid', 'value', 'timestamp']],
            name='expr',
            dshape="""var * {
                sid: int64,
                value: float64,
                timestamp: datetime,
            }""",
        )

        with self.assertRaises(TypeError) as e:
            from_blaze(
                expr,
                loader=self.garbage_loader,
                no_deltas_rule='ignore',
                no_checkpoints_rule='ignore',
            )
        self.assertIn("'asof_date'", str(e.exception))
        self.assertIn(repr(str(expr.dshape.measure)), str(e.exception))

    def test_missing_timestamp(self):
        expr = bz.data(
            self.df.loc[:, ['sid', 'value', 'asof_date']],
            name='expr',
            dshape="""var * {
                sid: int64,
                value: float64,
                asof_date: datetime,
            }""",
        )

        loader = BlazeLoader()

        ds = from_blaze(
            expr,
            loader=loader,
            no_deltas_rule='ignore',
            no_checkpoints_rule='ignore',
        )

        self.assertEqual(len(loader), 2)  # added the two columns
        for column in ds.columns:
            exprdata = loader[column]

            assert_isidentical(
                exprdata.expr,
                bz.transform(expr, timestamp=expr.asof_date),
            )

    def test_from_blaze_no_resources_dataset_expr(self):
        expr = bz.symbol('expr', self.dshape)

        with self.assertRaises(ValueError) as e:
            from_blaze(
                expr,
                loader=self.garbage_loader,
                no_deltas_rule='ignore',
                no_checkpoints_rule='ignore',
                missing_values=self.missing_values,
            )
        assert_equal(
            str(e.exception),
            'no resources provided to compute expr',
        )

    @parameter_space(metadata={'deltas', 'checkpoints'})
    def test_from_blaze_no_resources_metadata_expr(self, metadata):
        expr = bz.data(self.df, name='expr', dshape=self.dshape)
        metadata_expr = bz.symbol('metadata', self.dshape)

        with self.assertRaises(ValueError) as e:
            from_blaze(
                expr,
                loader=self.garbage_loader,
                no_deltas_rule='ignore',
                no_checkpoints_rule='ignore',
                missing_values=self.missing_values,
                **{metadata: metadata_expr}
            )
        assert_equal(
            str(e.exception),
            'no resources provided to compute %s' % metadata,
        )

    def test_from_blaze_mixed_resources_dataset_expr(self):
        expr = bz.data(self.df, name='expr', dshape=self.dshape)

        with self.assertRaises(ValueError) as e:
            from_blaze(
                expr,
                resources={expr: self.df},
                loader=self.garbage_loader,
                no_deltas_rule='ignore',
                no_checkpoints_rule='ignore',
                missing_values=self.missing_values,
            )
        assert_equal(
            str(e.exception),
            'explicit and implicit resources provided to compute expr',
        )

    @parameter_space(metadata={'deltas', 'checkpoints'})
    def test_from_blaze_mixed_resources_metadata_expr(self, metadata):
        expr = bz.symbol('expr', self.dshape)
        metadata_expr = bz.data(self.df, name=metadata, dshape=self.dshape)

        with self.assertRaises(ValueError) as e:
            from_blaze(
                expr,
                resources={metadata_expr: self.df},
                loader=self.garbage_loader,
                no_deltas_rule='ignore',
                no_checkpoints_rule='ignore',
                missing_values=self.missing_values,
                **{metadata: metadata_expr}
            )
        assert_equal(
            str(e.exception),
            'explicit and implicit resources provided to compute %s' %
            metadata,
        )

    @parameter_space(deltas={True, False}, checkpoints={True, False})
    def test_auto_metadata(self, deltas, checkpoints):
        select_level = op.getitem(('ignore', 'raise'))
        m = {'ds': self.df}
        if deltas:
            m['ds_deltas'] = pd.DataFrame(columns=self.df.columns),
        if checkpoints:
            m['ds_checkpoints'] = pd.DataFrame(columns=self.df.columns),
        expr = bz.data(
            m,
            dshape=var * Record((k, self.dshape.measure) for k in m),
        )
        loader = BlazeLoader()
        ds = from_blaze(
            expr.ds,
            loader=loader,
            missing_values=self.missing_values,
            no_deltas_rule=select_level(deltas),
            no_checkpoints_rule=select_level(checkpoints),
        )
        self.assertEqual(len(loader), 3)  # added the three columns
        for column in ds.columns:
            exprdata = loader[column]
            self.assertTrue(exprdata.expr.isidentical(expr.ds))
            if deltas:
                self.assertTrue(exprdata.deltas.isidentical(expr.ds_deltas))
            else:
                self.assertIsNone(exprdata.deltas)
            if checkpoints:
                self.assertTrue(
                    exprdata.checkpoints.isidentical(expr.ds_checkpoints),
                )
            else:
                self.assertIsNone(exprdata.checkpoints)

    @parameter_space(deltas={True, False}, checkpoints={True, False})
    def test_auto_metadata_fail_warn(self, deltas, checkpoints):
        select_level = op.getitem(('ignore', 'warn'))
        with warnings.catch_warnings(record=True) as ws:
            warnings.simplefilter('always')
            loader = BlazeLoader()
            expr = bz.data(self.df, dshape=self.dshape)
            from_blaze(
                expr,
                loader=loader,
                no_deltas_rule=select_level(deltas),
                no_checkpoints_rule=select_level(checkpoints),
                missing_values=self.missing_values,
            )
            self.assertEqual(len(ws), deltas + checkpoints)

        for w in ws:
            w = w.message
            self.assertIsInstance(w, NoMetaDataWarning)
            self.assertIn(str(expr), str(w))

    @parameter_space(deltas={True, False}, checkpoints={True, False})
    def test_auto_metadata_fail_raise(self, deltas, checkpoints):
        if not (deltas or checkpoints):
            # not a real case
            return
        select_level = op.getitem(('ignore', 'raise'))
        loader = BlazeLoader()
        expr = bz.data(self.df, dshape=self.dshape)
        with self.assertRaises(ValueError) as e:
            from_blaze(
                expr,
                loader=loader,
                no_deltas_rule=select_level(deltas),
                no_checkpoints_rule=select_level(checkpoints),
            )
        self.assertIn(str(expr), str(e.exception))

    def test_non_pipeline_field(self):
        expr = bz.data(
            [],
            dshape="""
            var * {
                 a: complex,
                 asof_date: datetime,
                 timestamp: datetime,
            }""",
        )
        ds = from_blaze(
            expr,
            loader=self.garbage_loader,
            no_deltas_rule='ignore',
            no_checkpoints_rule='ignore',
        )
        with self.assertRaises(AttributeError):
            ds.a
        self.assertIsInstance(
            object.__getattribute__(ds, 'a'),
            NonPipelineField,
        )

    @skipIf(new_pandas, skip_pipeline_new_pandas)
    def test_cols_with_all_missing_vals(self):
        """
        Tests that when there is no known data, we get output where the
        columns have the right dtypes and the right missing values filled in.

        input (self.df):
        Empty DataFrame
        Columns: [sid, float_value, str_value, int_value, bool_value, dt_value,
            asof_date, timestamp]
        Index: []

        output (expected)
                                          str_value  float_value  int_value
        2014-01-01 Equity(65 [A])      None          NaN          0
                   Equity(66 [B])      None          NaN          0
                   Equity(67 [C])      None          NaN          0
        2014-01-02 Equity(65 [A])      None          NaN          0
                   Equity(66 [B])      None          NaN          0
                   Equity(67 [C])      None          NaN          0
        2014-01-03 Equity(65 [A])      None          NaN          0
                   Equity(66 [B])      None          NaN          0
                   Equity(67 [C])      None          NaN          0

                                  dt_value  bool_value
        2014-01-01 Equity(65 [A])      NaT  False
                   Equity(66 [B])      NaT  False
                   Equity(67 [C])      NaT  False
        2014-01-02 Equity(65 [A])      NaT  False
                   Equity(66 [B])      NaT  False
                   Equity(67 [C])      NaT  False
        2014-01-03 Equity(65 [A])      NaT  False
                   Equity(66 [B])      NaT  False
                   Equity(67 [C])      NaT  False
        """
        df = empty_dataframe(
            ('sid', 'int64'),
            ('float_value', 'float64'),
            ('str_value', 'object'),
            ('int_value', 'int64'),
            ('bool_value', 'bool'),
            ('dt_value', 'datetime64[ns]'),
            ('asof_date', 'datetime64[ns]'),
            ('timestamp', 'datetime64[ns]'),
        )

        expr = bz.data(
            df,
            dshape="""
            var * {
                 sid: int64,
                 float_value: float64,
                 str_value: string,
                 int_value: int64,
                 bool_value: bool,
                 dt_value: datetime,
                 asof_date: datetime,
                 timestamp: datetime,
            }""",
        )
        fields = OrderedDict(expr.dshape.measure.fields)

        expected = pd.DataFrame({
            "str_value": np.array([None,
                                   None,
                                   None,
                                   None,
                                   None,
                                   None,
                                   None,
                                   None,
                                   None],
                                  dtype='object'),
            "float_value": np.array([np.NaN,
                                     np.NaN,
                                     np.NaN,
                                     np.NaN,
                                     np.NaN,
                                     np.NaN,
                                     np.NaN,
                                     np.NaN,
                                     np.NaN],
                                    dtype='float64'),
            "int_value": np.array([0,
                                   0,
                                   0,
                                   0,
                                   0,
                                   0,
                                   0,
                                   0,
                                   0],
                                  dtype='int64'),
            "bool_value": np.array([False,
                                    False,
                                    False,
                                    False,
                                    False,
                                    False,
                                    False,
                                    False,
                                    False],
                                   dtype='bool'),
            "dt_value": [pd.NaT,
                         pd.NaT,
                         pd.NaT,
                         pd.NaT,
                         pd.NaT,
                         pd.NaT,
                         pd.NaT,
                         pd.NaT,
                         pd.NaT],
        },
            columns=['str_value', 'float_value', 'int_value', 'bool_value',
                     'dt_value'],
            index=pd.MultiIndex.from_product(
                (self.dates.tz_localize('UTC'), self.asset_finder.retrieve_all(
                    self.ASSET_FINDER_EQUITY_SIDS
                ))
            )
        )

        self._test_id(
            df,
            var * Record(fields),
            expected,
            self.asset_finder,
            ('float_value', 'str_value', 'int_value', 'bool_value',
             'dt_value'),
        )

    @skipIf(new_pandas, skip_pipeline_new_pandas)
    def test_cols_with_some_missing_vals(self):
        """
        Tests the following:
            1) Forward filling replaces missing values correctly for the data
            types supported in pipeline.
            2) We don't forward fill when the missing value is the actual value
             we got for a date in the case of int/bool columns.
            3) We get the correct type of missing value in the output.

        input (self.df):
           asof_date bool_value   dt_value  float_value  int_value  sid
        0 2014-01-01       True 2011-01-01            0          1   65
        1 2014-01-03       True 2011-01-02            1          2   66
        2 2014-01-01       True 2011-01-03            2          3   67
        3 2014-01-02      False        NaT          NaN          0   67

          str_value  timestamp
        0         a  2014-01-01
        1         b  2014-01-03
        2         c  2014-01-01
        3      None  2014-01-02

        output (expected)
                                  str_value  float_value  int_value bool_value
        2014-01-01 Equity(65 [A])         a            0          1       True
                   Equity(66 [B])      None          NaN          0      False
                   Equity(67 [C])         c            2          3       True
        2014-01-02 Equity(65 [A])         a            0          1       True
                   Equity(66 [B])      None          NaN          0      False
                   Equity(67 [C])         c            2          0      False
        2014-01-03 Equity(65 [A])         a            0          1       True
                   Equity(66 [B])         b            1          2       True
                   Equity(67 [C])         c            2          0      False

                                    dt_value
        2014-01-01 Equity(65 [A]) 2011-01-01
                   Equity(66 [B])        NaT
                   Equity(67 [C]) 2011-01-03
        2014-01-02 Equity(65 [A]) 2011-01-01
                   Equity(66 [B])        NaT
                   Equity(67 [C]) 2011-01-03
        2014-01-03 Equity(65 [A]) 2011-01-01
                   Equity(66 [B]) 2011-01-02
                   Equity(67 [C]) 2011-01-03
        """
        dates = pd.Index([
            self.dates[0],
            self.dates[-1],
            self.dates[0],
            self.dates[1],
        ])
        df = pd.DataFrame({
            'sid': self.ASSET_FINDER_EQUITY_SIDS[:-1] +
            (self.ASSET_FINDER_EQUITY_SIDS[-1],) * 2,
            'float_value': (0., 1., 2., np.NaN),
            'str_value': ('a', 'b', 'c', None),
            'cat_value': pd.Categorical(
                values=['a', 'b', 'c', None],
                categories=['a', 'b', 'c', None],
            ),
            'int_value': (1, 2, 3, 0),
            'bool_value': (True, True, True, False),
            'dt_value': (pd.Timestamp('2011-01-01'),
                         pd.Timestamp('2011-01-02'),
                         pd.Timestamp('2011-01-03'),
                         pd.NaT),
            'asof_date': dates - pd.Timedelta(days=2),
            'timestamp': dates - pd.Timedelta(days=1),
        })

        expr = bz.data(
            df,
            dshape="""
            var * {
                 sid: int64,
                 float_value: float64,
                 str_value: string,
                 cat_value: string,
                 int_value: int64,
                 bool_value: bool,
                 dt_value: datetime,
                 asof_date: datetime,
                 timestamp: datetime,
            }""",
        )
        fields = OrderedDict(expr.dshape.measure.fields)

        expected = pd.DataFrame(
            {
                'str_value': np.array(
                    ['a', None, 'c', 'a', None, 'c', 'a', 'b', 'c'],
                    dtype='object',
                ),
                'cat_value': np.array(
                    ['a', None, 'c', 'a', None, 'c', 'a', 'b', 'c'],
                    dtype='object',
                ),
                'float_value': np.array(
                    [0, np.NaN, 2, 0, np.NaN, 2, 0, 1, 2],
                    dtype='float64',
                ),
                'int_value': np.array(
                    [1, 0, 3, 1, 0, 3, 1, 2, 3],
                    dtype='int64',
                ),
                'bool_value': np.array(
                    [True, False, True, True, False, False, True, True, False],
                    dtype='bool',
                ),
                'dt_value': [
                    pd.Timestamp('2011-01-01'),
                    pd.NaT,
                    pd.Timestamp('2011-01-03'),
                    pd.Timestamp('2011-01-01'),
                    pd.NaT,
                    pd.Timestamp('2011-01-03'),
                    pd.Timestamp('2011-01-01'),
                    pd.Timestamp('2011-01-02'),
                    pd.Timestamp('2011-01-03'),
                ],
            },
            columns=[
                'str_value',
                'cat_value',
                'float_value',
                'int_value',
                'bool_value',
                'dt_value',
            ],
            index=pd.MultiIndex.from_product((
                self.dates.tz_localize('UTC'),
                self.asset_finder.retrieve_all(self.ASSET_FINDER_EQUITY_SIDS),
            )),
        )

        self._test_id(
            df,
            var * Record(fields),
            expected,
            self.asset_finder,
            expected.columns,
        )

    def test_complex_expr(self):
        expr = bz.data(self.df, dshape=self.dshape, name='expr')
        # put an Add in the table
        expr_with_add = bz.transform(expr, value=expr.value + 1)

        # test that we can have complex expressions with no metadata
        from_blaze(
            expr_with_add,
            deltas=None,
            checkpoints=None,
            loader=self.garbage_loader,
            missing_values=self.missing_values,
            no_checkpoints_rule='ignore',
        )

        with self.assertRaises(TypeError) as e:
            # test that we cannot create a single column from a non field
            from_blaze(
                expr.value + 1,  # put an Add in the column
                deltas=None,
                checkpoints=None,
                loader=self.garbage_loader,
                missing_values=self.missing_values,
                no_checkpoints_rule='ignore',
            )
        assert_equal(
            str(e.exception),
            "expression 'expr.value + 1' was array-like but not a simple field"
            " of some larger table",
        )

        deltas = bz.data(
            pd.DataFrame(columns=self.df.columns),
            dshape=self.dshape,
            name='deltas',
        )
        checkpoints = bz.data(
            pd.DataFrame(columns=self.df.columns),
            dshape=self.dshape,
            name='checkpoints',
        )

        # test that we can have complex expressions with explicit metadata
        from_blaze(
            expr_with_add,
            deltas=deltas,
            checkpoints=checkpoints,
            loader=self.garbage_loader,
            missing_values=self.missing_values,
        )

        with self.assertRaises(TypeError) as e:
            # test that we cannot create a single column from a non field
            # even with explicit metadata
            from_blaze(
                expr.value + 1,
                deltas=deltas,
                checkpoints=checkpoints,
                loader=self.garbage_loader,
                missing_values=self.missing_values,
            )
        assert_equal(
            str(e.exception),
            "expression 'expr.value + 1' was array-like but not a simple field"
            " of some larger table",
        )

    def _test_id(self, df, dshape, expected, finder, add):
        expr = bz.data(df, name='expr', dshape=dshape)
        loader = BlazeLoader()
        domain = self.create_domain(self.dates)
        ds = from_blaze(
            expr,
            loader=loader,
            no_deltas_rule='ignore',
            no_checkpoints_rule='ignore',
            missing_values=self.missing_values,
            domain=domain
        )
        p = Pipeline(domain=domain)
        for a in add:
            p.add(getattr(ds, a).latest, a)
        dates = self.dates

        result = SimplePipelineEngine(
            loader, finder,
        ).run_pipeline(p, dates[0], dates[-1])
        assert_frame_equal(
            result.sort_index(axis=1),
            expected.sort_index(axis=1),
            check_dtype=False,
        )

    def _test_id_macro(self, df, dshape, expected, finder, add, dates=None):
        if dates is None:
            dates = self.dates
        expr = bz.data(df, name='expr', dshape=dshape)
        loader = BlazeLoader()
        domain = self.create_domain(dates)
        ds = from_blaze(
            expr,
            loader=loader,
            no_deltas_rule='ignore',
            no_checkpoints_rule='ignore',
            missing_values=self.missing_values,
            domain=domain,
        )

        p = Pipeline(domain=domain)
        macro_inputs = []
        for column_name in add:
            column = getattr(ds, column_name)
            macro_inputs.append(column)
            with self.assertRaises(UnsupportedPipelineOutput):
                # Single column output terms cannot be added to a pipeline.
                p.add(column.latest, column_name)

        class UsesMacroInputs(CustomFactor):
            inputs = macro_inputs
            window_length = 1

            def compute(self, today, assets, out, *inputs):
                e = expected.loc[today]
                for i, input_ in enumerate(inputs):
                    # Each macro input should only have one column.
                    assert_equal(input_.shape, (self.window_length, 1))
                    assert_equal(input_[0, 0], e[i])

        # Run the pipeline with our custom factor. Assertions about the
        # expected macro data are made in the `compute` function of our custom
        # factor above.
        p.add(UsesMacroInputs(), 'uses_macro_inputs')
        engine = SimplePipelineEngine(loader, finder)
        engine.run_pipeline(p, dates[0], dates[-1])

    def test_custom_query_time_tz(self):
        """
        input (df):
           asof_date  int_value  sid           timestamp  value
        0 2013-12-31          0   65 2014-01-01 13:44:00    0.0
        1 2013-12-31          1   66 2014-01-01 13:44:00    1.0
        2 2013-12-31          2   67 2014-01-01 13:44:00    2.0
        3 2013-12-31          1   65 2014-01-01 13:45:00    1.0
        4 2013-12-31          2   66 2014-01-01 13:45:00    2.0
        5 2013-12-31          3   67 2014-01-01 13:45:00    3.0
        6 2014-01-02          2   65 2014-01-03 13:44:00    2.0
        7 2014-01-02          3   66 2014-01-03 13:44:00    3.0
        8 2014-01-02          4   67 2014-01-03 13:44:00    4.0

        output (expected):
                                                  int_value  value
        2014-01-01 00:00:00+00:00 Equity(65 [A])          0    0.0
                                  Equity(66 [B])          1    1.0
                                  Equity(67 [C])          2    2.0
        2014-01-02 00:00:00+00:00 Equity(65 [A])          1    1.0
                                  Equity(66 [B])          2    2.0
                                  Equity(67 [C])          3    3.0
        2014-01-03 00:00:00+00:00 Equity(65 [A])          2    2.0
                                  Equity(66 [B])          3    3.0
                                  Equity(67 [C])          4    4.0
        """
        df = self.df.copy()
        df['timestamp'] = (
            pd.DatetimeIndex(df['timestamp'], tz='EST') +
            timedelta(hours=8, minutes=44)
        ).tz_convert('utc').tz_localize(None)
        df.ix[3:5, 'timestamp'] = pd.Timestamp('2014-01-01 13:45')
        expr = bz.data(df, name='expr', dshape=self.dshape)
        loader = BlazeLoader()
        ds = from_blaze(
            expr,
            loader=loader,
            no_deltas_rule='ignore',
            no_checkpoints_rule='ignore',
            missing_values=self.missing_values,
            domain=self.create_domain(
                self.dates,
                data_query_time=time(8, 45, tzinfo=pytz.timezone('EST')),
            ),
        )
        p = Pipeline()
        p.add(ds.value.latest, 'value')
        p.add(ds.int_value.latest, 'int_value')

        result = SimplePipelineEngine(
            loader, self.asset_finder,
        ).run_pipeline(p, self.dates[0], self.dates[-1])

        expected = df.drop('asof_date', axis=1)
        expected['timestamp'] = expected['timestamp'].dt.normalize().astype(
            'datetime64[ns]',
        ).dt.tz_localize('utc')
        expected.ix[3:5, 'timestamp'] += timedelta(days=1)
        expected.set_index(['timestamp', 'sid'], inplace=True)
        expected.index = pd.MultiIndex.from_product((
            expected.index.levels[0],
            self.asset_finder.retrieve_all(expected.index.levels[1]),
        ))
        assert_frame_equal(result, expected, check_dtype=False)

    def test_id(self):
        """
        input (self.df):
           asof_date  sid  timestamp int_value value
        0 2014-01-01   65 2014-01-01         0     0
        1 2014-01-01   66 2014-01-01         1     1
        2 2014-01-01   67 2014-01-01         2     2
        3 2014-01-02   65 2014-01-02         1     1
        4 2014-01-02   66 2014-01-02         2     2
        5 2014-01-02   67 2014-01-02         3     3
        6 2014-01-03   65 2014-01-03         2     2
        7 2014-01-03   66 2014-01-03         3     3
        8 2014-01-03   67 2014-01-03         4     4

        output (expected)
                                  int_value value
        2014-01-01 Equity(65 [A])         0     0
                   Equity(66 [B])         1     1
                   Equity(67 [C])         2     2
        2014-01-02 Equity(65 [A])         1     1
                   Equity(66 [B])         2     2
                   Equity(67 [C])         3     3
        2014-01-03 Equity(65 [A])         2     2
                   Equity(66 [B])         3     3
                   Equity(67 [C])         4     4
        """
        expected = self.df.drop(['timestamp', 'asof_date', 'sid'], axis=1)
        expected.index = pd.MultiIndex.from_product((
            self.dates.tz_localize('UTC'),
            self.asset_finder.retrieve_all(self.asset_finder.sids),
        ))
        self._test_id(
            self.df,
            self.dshape,
            expected,
            self.asset_finder,
            ('int_value', 'value',)
        )

    def test_id_with_asof_date(self):
        """
        input (self.df):
           asof_date  sid  timestamp int_value value
        0 2014-01-01   65 2014-01-01         0     0
        1 2014-01-01   66 2014-01-01         1     1
        2 2014-01-01   67 2014-01-01         2     2
        3 2014-01-02   65 2014-01-02         1     1
        4 2014-01-02   66 2014-01-02         2     2
        5 2014-01-02   67 2014-01-02         3     3
        6 2014-01-03   65 2014-01-03         2     2
        7 2014-01-03   66 2014-01-03         3     3
        8 2014-01-03   67 2014-01-03         4     4

        output (expected)
                                    asof_date
        2014-01-01 Equity(65 [A])  2014-01-01
                   Equity(66 [B])  2014-01-01
                   Equity(67 [C])  2014-01-01
        2014-01-02 Equity(65 [A])  2014-01-02
                   Equity(66 [B])  2014-01-02
                   Equity(67 [C])  2014-01-02
        2014-01-03 Equity(65 [A])  2014-01-03
                   Equity(66 [B])  2014-01-03
                   Equity(67 [C])  2014-01-03
        """
        expected = self.df.drop(
            ['timestamp', 'sid', 'value', 'int_value'],
            axis=1,
        )
        expected.index = pd.MultiIndex.from_product((
            self.dates.tz_localize('UTC'),
            self.asset_finder.retrieve_all(self.asset_finder.sids),
        ))
        self._test_id(
            self.df,
            self.dshape,
            expected,
            self.asset_finder,
            ('asof_date',)
        )

    def test_id_ffill_out_of_window(self):
        """
        input (df):

           asof_date  timestamp  sid  other  value
        0 2013-12-22 2013-12-22   65      0      0
        1 2013-12-22 2013-12-22   66    NaN      1
        2 2013-12-22 2013-12-22   67      2    NaN
        3 2013-12-23 2013-12-23   65    NaN      1
        4 2013-12-23 2013-12-23   66      2    NaN
        5 2013-12-23 2013-12-23   67      3      3
        6 2013-12-24 2013-12-24   65      2    NaN
        7 2013-12-24 2013-12-24   66      3      3
        8 2013-12-24 2013-12-24   67    NaN      4

        output (expected):
                                   other  value
        2014-01-01 Equity(65 [A])      2      1
                   Equity(66 [B])      3      3
                   Equity(67 [C])      3      4
        2014-01-02 Equity(65 [A])      2      1
                   Equity(66 [B])      3      3
                   Equity(67 [C])      3      4
        2014-01-03 Equity(65 [A])      2      1
                   Equity(66 [B])      3      3
                   Equity(67 [C])      3      4
        """
        dates = self.dates.repeat(3) - timedelta(days=10)
        df = pd.DataFrame({
            'sid': self.ASSET_FINDER_EQUITY_SIDS * 3,
            'value': (0, 1, np.nan, 1, np.nan, 3, np.nan, 3, 4),
            'other': (0, np.nan, 2, np.nan, 2, 3, 2, 3, np.nan),
            'asof_date': dates,
            'timestamp': dates,
        })
        fields = OrderedDict(self.dshape.measure.fields)
        fields['other'] = fields['value']

        expected = pd.DataFrame(
            np.array([[2, 1],
                      [3, 3],
                      [3, 4],
                      [2, 1],
                      [3, 3],
                      [3, 4],
                      [2, 1],
                      [3, 3],
                      [3, 4]]),
            columns=['other', 'value'],
            index=pd.MultiIndex.from_product(
                (self.dates.tz_localize('UTC'), self.asset_finder.retrieve_all(
                    self.ASSET_FINDER_EQUITY_SIDS
                )),
            ),
        )
        self._test_id(
            df,
            var * Record(fields),
            expected,
            self.asset_finder,
            ('value', 'other'),
        )

    def test_id_multiple_columns(self):
        """
        input (df):
           asof_date  sid  timestamp  value  other
        0 2014-01-01   65 2014-01-01      0      1
        1 2014-01-01   66 2014-01-01      1      2
        2 2014-01-01   67 2014-01-01      2      3
        3 2014-01-02   65 2014-01-02      1      2
        4 2014-01-02   66 2014-01-02      2      3
        5 2014-01-02   67 2014-01-02      3      4
        6 2014-01-03   65 2014-01-03      2      3
        7 2014-01-03   66 2014-01-03      3      4
        8 2014-01-03   67 2014-01-03      4      5

        output (expected):
                                   value  other
        2014-01-01 Equity(65 [A])      0      1
                   Equity(66 [B])      1      2
                   Equity(67 [C])      2      3
        2014-01-02 Equity(65 [A])      1      2
                   Equity(66 [B])      2      3
                   Equity(67 [C])      3      4
        2014-01-03 Equity(65 [A])      2      3
                   Equity(66 [B])      3      4
                   Equity(67 [C])      4      5
        """
        df = self.df.copy()
        df['other'] = df.value + 1
        fields = OrderedDict(self.dshape.measure.fields)
        fields['other'] = fields['value']

        expected = df.drop(['timestamp', 'asof_date', 'sid'], axis=1)
        expected.index = pd.MultiIndex.from_product((
            self.dates.tz_localize('UTC'),
            self.asset_finder.retrieve_all(self.asset_finder.sids),
        ))
        self._test_id(
            df,
            var * Record(fields),
            expected,
            self.asset_finder,
            ('value', 'int_value', 'other'),
        )

    def test_id_macro_dataset(self):
        """
        input (self.macro_df)
           asof_date  timestamp  value
        0 2014-01-01 2014-01-01      0
        3 2014-01-02 2014-01-02      1
        6 2014-01-03 2014-01-03      2

        output (expected):
                    value
        2014-01-01      0
        2014-01-02      1
        2014-01-03      2
        """
        expected = pd.DataFrame(
            data=[[0],
                  [1],
                  [2]],
            columns=['value'],
            index=self.dates,
        )
        self._test_id_macro(
            self.macro_df,
            self.macro_dshape,
            expected,
            self.asset_finder,
            ('value',),
        )

    def test_id_ffill_out_of_window_macro_dataset(self):
        """
        input (df):
           asof_date  timestamp  other  value
        0 2013-12-22 2013-12-22    NaN      0
        1 2013-12-23 2013-12-23      1    NaN
        2 2013-12-24 2013-12-24    NaN    NaN

        output (expected):
                    other  value
        2014-01-01      1      0
        2014-01-02      1      0
        2014-01-03      1      0
        """
        dates = self.dates - timedelta(days=10)
        df = pd.DataFrame({
            'value': (0, np.nan, np.nan),
            'other': (np.nan, 1, np.nan),
            'asof_date': dates,
            'timestamp': dates,
        })
        fields = OrderedDict(self.macro_dshape.measure.fields)
        fields['other'] = fields['value']

        expected = pd.DataFrame(
            data=[[0, 1],
                  [0, 1],
                  [0, 1]],
            columns=['other', 'value'],
            index=self.dates.tz_localize('UTC'),
        )
        self._test_id_macro(
            df,
            var * Record(fields),
            expected,
            self.asset_finder,
            ('value', 'other'),
        )

    def test_id_macro_dataset_multiple_columns(self):
        """
        input (df):
           asof_date  timestamp  other  value
        0 2014-01-01 2014-01-01      1      0
        3 2014-01-02 2014-01-02      2      1
        6 2014-01-03 2014-01-03      3      2

        output (expected):
                    other  value
        2014-01-01      1      0
        2014-01-02      2      1
        2014-01-03      3      2
        """
        df = self.macro_df.copy()
        df['other'] = df.value + 1
        fields = OrderedDict(self.macro_dshape.measure.fields)
        fields['other'] = fields['value']

        with tmp_asset_finder(equities=simple_asset_info) as finder:
            expected = pd.DataFrame(
                data=[[0, 1],
                      [1, 2],
                      [2, 3]],
                columns=['value', 'other'],
                index=self.dates,
                dtype=np.float64,
            )
            self._test_id_macro(
                df,
                var * Record(fields),
                expected,
                finder,
                ('value', 'other'),
            )

    def test_id_take_last_in_group(self):
        T = pd.Timestamp
        df = pd.DataFrame(
            columns=['asof_date',        'timestamp', 'sid', 'other', 'value'],
            data=[
                [T('2013-12-31'), T('2013-12-31 22'),    65,        0,      0],
                [T('2013-12-31'), T('2013-12-31 23'),    65,        1, np.nan],
                [T('2013-12-31'), T('2013-12-31 22'),    66,   np.nan, np.nan],
                [T('2013-12-31'), T('2013-12-31 23'),    66,   np.nan,      1],
                [T('2013-12-31'), T('2013-12-31 22'),    67,        2, np.nan],
                [T('2013-12-31'), T('2013-12-31 23'),    67,   np.nan, np.nan],
                [T('2014-01-01'), T('2014-01-01 22'),    65,   np.nan, np.nan],
                [T('2014-01-01'), T('2014-01-01 23'),    65,   np.nan,      1],
                [T('2014-01-01'), T('2014-01-01 22'),    66,   np.nan, np.nan],
                [T('2014-01-01'), T('2014-01-01 23'),    66,        2, np.nan],
                [T('2014-01-01'), T('2014-01-01 22'),    67,        3,      3],
                [T('2014-01-01'), T('2014-01-01 23'),    67,        3,      3],
                [T('2014-01-02'), T('2014-01-02 22'),    65,        2, np.nan],
                [T('2014-01-02'), T('2014-01-02 23'),    65,        2, np.nan],
                [T('2014-01-02'), T('2014-01-02 22'),    66,        3,      3],
                [T('2014-01-02'), T('2014-01-02 23'),    66,   np.nan, np.nan],
                [T('2014-01-02'), T('2014-01-02 22'),    67,   np.nan, np.nan],
                [T('2014-01-02'), T('2014-01-02 23'),    67,   np.nan,      4],
            ],
        )
        fields = OrderedDict(self.dshape.measure.fields)
        fields['other'] = fields['value']

        expected = pd.DataFrame(
            columns=['other', 'value'],
            data=[
                [1,           0],  # 2014-01-01 Equity(65 [A])
                [np.nan,      1],             # Equity(66 [B])
                [2,      np.nan],             # Equity(67 [C])
                [1,           1],  # 2014-01-02 Equity(65 [A])
                [2,           1],             # Equity(66 [B])
                [3,           3],             # Equity(67 [C])
                [2,           1],  # 2014-01-03 Equity(65 [A])
                [3,           3],             # Equity(66 [B])
                [3,           4],             # Equity(67 [C])
            ],
            index=pd.MultiIndex.from_product(
                (self.dates.tz_localize('UTC'), self.asset_finder.retrieve_all(
                    self.ASSET_FINDER_EQUITY_SIDS
                )),
            ),
        )
        self._test_id(
            df,
            var * Record(fields),
            expected,
            self.asset_finder,
            ('value', 'other'),
        )

    def test_id_take_last_in_group_macro(self):
        """
        output (expected):

                    other  value
        2014-01-01    NaN      2
        2014-01-02      1      3
        2014-01-03      2      3
         """
        T = pd.Timestamp
        df = pd.DataFrame(
            columns=['asof_date',        'timestamp', 'other', 'value'],
            data=[
                [T('2013-12-31'), T('2013-12-31 01'),   np.nan,      1],
                [T('2013-12-31'), T('2013-12-31 02'),   np.nan,      2],
                [T('2014-01-01'), T('2014-01-01 01'),        1, np.nan],
                [T('2014-01-01'), T('2014-01-01 02'),   np.nan,      3],
                [T('2014-01-02'), T('2014-01-02 01'),        2, np.nan],
                [T('2014-01-02'), T('2014-01-02 02'),   np.nan, np.nan],
            ],
        )
        fields = OrderedDict(self.macro_dshape.measure.fields)
        fields['other'] = fields['value']

        expected = pd.DataFrame(
            data=[[np.nan, 2],   # 2014-01-01
                  [1,      3],   # 2014-01-02
                  [2,      3]],  # 2014-01-03
            columns=['other', 'value'],
            index=self.dates,
        )
        self._test_id_macro(
            df,
            var * Record(fields),
            expected,
            self.asset_finder,
            ('other', 'value'),
        )

    def _run_pipeline(self,
                      expr,
                      deltas,
                      checkpoints,
                      expected_views,
                      expected_output,
                      finder,
                      calendar,
                      start,
                      end,
                      window_length,
                      compute_fn=None):
        loader = BlazeLoader()
        ds = from_blaze(
            expr,
            deltas,
            checkpoints,
            loader=loader,
            no_deltas_rule='raise',
            no_checkpoints_rule='ignore',
            missing_values=self.missing_values,
            domain=self.create_domain(calendar),
        )
        p = Pipeline()

        # prevent unbound locals issue in the inner class
        window_length_ = window_length

        if compute_fn is None:
            self.assertIsNone(
                expected_output,
                'expected_output must be None if compute_fn is None',
            )

            def compute_fn(data):
                return data[0]

        class TestFactor(CustomFactor):
            inputs = ds.value,
            window_length = window_length_

            def compute(self, today, assets, out, data):
                assert_array_almost_equal(
                    data,
                    expected_views[today],
                    err_msg=str(today),
                )
                out[:] = compute_fn(data)

        p.add(TestFactor(), 'value')

        result = SimplePipelineEngine(
            loader, finder,
        ).run_pipeline(p, start, end)

        if expected_output is not None:
            assert_frame_equal(
                result,
                expected_output,
                check_dtype=False,
            )

    @with_ignore_sid()
    def test_deltas(self, asset_info, add_extra_sid):
        df = self.df.copy()
        if add_extra_sid:
            extra_sid_df = pd.DataFrame({
                'asof_date': self.asof_dates,
                'timestamp': self.timestamps,
                'sid': (ord('E'),) * 3,
                'value': (3., 4., 5.,),
                'int_value': (3, 4, 5),
            })
            df = df.append(extra_sid_df, ignore_index=True)
        expr = bz.data(df, name='expr', dshape=self.dshape)
        deltas = bz.data(df, dshape=self.dshape)
        deltas = bz.data(
            odo(
                bz.transform(
                    deltas,
                    value=deltas.value + 10,
                    timestamp=deltas.timestamp + timedelta(days=1),
                ),
                pd.DataFrame,
            ),
            name='delta',
            dshape=self.dshape,
        )

        expected_views = keymap(partial(pd.Timestamp, tz='UTC'), {
            '2014-01-02': np.array([[10.0, 11.0, 12.0],
                                    [1.0, 2.0, 3.0]]),
            '2014-01-03': np.array([[11.0, 12.0, 13.0],
                                    [2.0, 3.0, 4.0]]),
            '2014-01-04': np.array([[12.0, 13.0, 14.0],
                                    [12.0, 13.0, 14.0]]),
        })

        nassets = len(asset_info)
        if nassets == 4:
            expected_views = valmap(
                lambda view: np.c_[view, [np.nan, np.nan]],
                expected_views,
            )
        with tmp_asset_finder(equities=asset_info) as finder:
            expected_output = pd.DataFrame(
                list(concatv([12] * nassets, [13] * nassets, [14] * nassets)),
                index=pd.MultiIndex.from_product((
                    sorted(expected_views.keys()),
                    finder.retrieve_all(asset_info.index),
                )),
                columns=('value',),
            )
            dates = self.dates
            dates = dates.insert(len(dates), dates[-1] + timedelta(days=1))
            self._run_pipeline(
                expr,
                deltas,
                None,
                expected_views,
                expected_output,
                finder,
                calendar=dates,
                start=dates[1],
                end=dates[-1],
                window_length=2,
                compute_fn=np.nanmax,
            )

    @with_ignore_sid()
    def test_deltas_before_index_0(self, asset_info, add_extra_sid):
        df = empty_dataframe(
            ('sid', 'int64'),
            ('value', 'float64'),
            ('asof_date', 'datetime64[ns]'),
            ('timestamp', 'datetime64[ns]'),
        )
        expr = bz.data(df, name='expr', dshape=self.dshape)

        T = pd.Timestamp
        # These data are interesting because we have four rows with an asof
        # date prior to the start of the query window. The first, second, and
        # fourth rows should become the best-known value on their timestamp.
        # The third row's asof date is less than the second row's asof date so,
        # due to forward filling rules, it is *not* the most recent value on
        # its timestamp. The value for row three should never be shown to the
        # user.
        deltas_df_single_sid = pd.DataFrame({
            'value': [0.0, 1.0, 2.0, 3.0],
            'asof_date': [
                T('2013-12-01'),
                T('2013-12-15'),
                T('2013-12-02'),  # not more recent than the previous day
                T('2013-12-16'),
            ],
            'timestamp': [
                T('2014-01-01 23:00'),
                T('2014-01-02 23:00'),
                T('2014-01-03 23:00'),
                T('2014-01-04 23:00'),
            ],
        })
        sids = asset_info.index
        if add_extra_sid:
            # add a sid to the dataset that the asset finder doesn't know about
            sids = sids.insert(0, ord('Z'))

        deltas_df = pd.concat([
            deltas_df_single_sid.assign(
                sid=sid,
                value=deltas_df_single_sid.value + (100 * n),
            )
            for n, sid in enumerate(asset_info.index)
        ])
        deltas = bz.data(deltas_df, name='deltas', dshape=self.dshape)

        expected_views_single_sid = keymap(partial(pd.Timestamp, tz='UTC'), {
            '2014-01-02': np.array([[0.0],
                                    [0.0]]),
            '2014-01-03': np.array([[1.0],
                                    [1.0]]),
            # The third row's value of 2.0 is *not* the best known value
            # because its asof date of 2013-12-02 is earlier than the previous
            # row's asof date of 2013-12-15. We continue to surface the second
            # row's value on this day.
            '2014-01-04': np.array([[1.0],
                                    [1.0]]),
            '2014-01-05': np.array([[3.0],
                                    [3.0]]),
        })

        column_constant = np.arange(len(asset_info)) * 100
        expected_views = {
            k: v + column_constant
            for k, v in expected_views_single_sid.items()
        }
        with tmp_asset_finder(equities=asset_info) as finder:
            dates = pd.date_range('2014-01-01', '2014-01-05')
            self._run_pipeline(
                expr,
                deltas,
                None,
                expected_views,
                None,
                finder,
                calendar=dates,
                start=dates[1],
                end=dates[-1],
                window_length=2,
            )

    @with_ignore_sid()
    def test_deltas_on_same_ix_out_of_order(self, asset_info, add_extra_sid):
        df = empty_dataframe(
            ('sid', 'int64'),
            ('value', 'float64'),
            ('asof_date', 'datetime64[ns]'),
            ('timestamp', 'datetime64[ns]'),
        )
        expr = bz.data(df, name='expr', dshape=self.dshape)

        T = pd.Timestamp

        # These data are interesting because we have pairs of rows that come on
        # the same asof_date in index space. The catch is that the asof dates
        # are sometimes out of order relative to their timestamps. This is used
        # to test cases where we get novel rows for dates between trading days
        # (weekends and holidays) although we learn about them out of order.
        #
        # The first two rows both map to index 0 in the output. The first row
        # has an earlier timestamp but later asof date so it should be
        # selected.
        #
        # The third and fourth rows both map to index 1 in the output. The
        # fourth row (second in the group) has both a later timestamp and asof
        # date so it should be selected.
        #
        # The fifth and sixth rows both map to index 2 in the output. The fifth
        # row (first in the group) has an earlier timestamp but later asof date
        # so it should be selected.
        deltas_df_single_sid = pd.DataFrame({
            'value': [
                0.0,  # selected
                1.0,  # ignored

                2.0,  # ignored
                3.0,  # selected

                4.0,  # selected
                5.0,  # ignored
            ],
            'asof_date': [
                # swapped order: second row is before the first
                T('2014-01-02'),
                T('2014-01-01'),

                # chronological order: second row is after the first
                T('2014-01-03'),
                T('2014-01-04'),

                # swapped order: second row is before the first
                T('2014-01-06'),
                T('2014-01-05'),
            ],
            'timestamp': [
                # we learn about all rows in monotonically increasing order
                T('2013-01-02 22:00'),
                T('2014-01-02 23:00'),
                T('2014-01-04 22:00'),
                T('2014-01-04 23:00'),
                T('2014-01-06 22:00'),
                T('2014-01-06 23:00'),
            ],
        })
        sids = asset_info.index
        if add_extra_sid:
            # add a sid to the dataset that the asset finder doesn't know about
            sids = sids.insert(0, ord('Z'))

        deltas_df = pd.concat([
            deltas_df_single_sid.assign(
                sid=sid,
                value=deltas_df_single_sid.value + (100 * n),
            )
            for n, sid in enumerate(asset_info.index)
        ])
        deltas = bz.data(deltas_df, name='deltas', dshape=self.dshape)

        expected_views_single_sid = keymap(partial(pd.Timestamp, tz='UTC'), {
            '2014-01-05': np.array([[0.0],
                                    [3.0]]),
            '2014-01-07': np.array([[3.0],
                                    [4.0]]),
        })

        column_constant = np.arange(len(asset_info)) * 100
        expected_views = {
            k: v + column_constant
            for k, v in expected_views_single_sid.items()
        }
        with tmp_asset_finder(equities=asset_info) as finder:
            # The dates queried are non-contiguous. We have two day groups to
            # capture the two day pairs in the input data.
            dates = pd.to_datetime(['2014-01-03', '2014-01-05', '2014-01-07'])
            self._run_pipeline(
                expr=expr,
                deltas=deltas,
                checkpoints=None,
                expected_views=expected_views,
                expected_output=None,
                finder=finder,
                calendar=dates,
                start=dates[1],
                end=dates[-1],
                window_length=2,
            )

    @with_extra_sid()
    def test_deltas_only_one_delta_in_universe(self, asset_info):
        expr = bz.data(self.df, name='expr', dshape=self.dshape)
        deltas = pd.DataFrame({
            'sid': [65, 66],
            'asof_date': [self.asof_dates[1], self.asof_dates[0]],
            'timestamp': [self.timestamps[2], self.timestamps[1]],
            'value': [10, 11],
        })
        deltas = bz.data(deltas, name='deltas', dshape=self.dshape)

        expected_views = keymap(partial(pd.Timestamp, tz='UTC'), {
            '2014-01-02': np.array([[0.0, 11.0, 2.0],
                                    [1.0, 2.0, 3.0]]),
            '2014-01-03': np.array([[10.0, 2.0, 3.0],
                                    [2.0, 3.0, 4.0]]),
            '2014-01-04': np.array([[2.0, 3.0, 4.0],
                                    [2.0, 3.0, 4.0]]),
        })

        nassets = len(asset_info)
        if nassets == 4:
            expected_views = valmap(
                lambda view: np.c_[view, [np.nan, np.nan]],
                expected_views,
            )

        with tmp_asset_finder(equities=asset_info) as finder:
            expected_output = pd.DataFrame(
                columns=[
                    'value',
                ],
                data=np.array([11, 10, 4]).repeat(len(asset_info.index)),
                index=pd.MultiIndex.from_product((
                    sorted(expected_views.keys()),
                    finder.retrieve_all(asset_info.index),
                )),
            )
            dates = self.dates
            dates = dates.insert(len(dates), dates[-1] + timedelta(days=1))
            self._run_pipeline(
                expr,
                deltas,
                None,
                expected_views,
                expected_output,
                finder,
                calendar=dates,
                start=dates[1],
                end=dates[-1],
                window_length=2,
                compute_fn=np.nanmax,
            )

    def test_deltas_macro(self):
        expr = bz.data(self.macro_df, name='expr', dshape=self.macro_dshape)
        deltas = bz.data(
            self.macro_df.iloc[:-1],
            name='deltas',
            dshape=self.macro_dshape,
        )
        deltas = bz.transform(
            deltas,
            value=deltas.value + 10,
            timestamp=deltas.timestamp + timedelta(days=1),
        )

        nassets = len(simple_asset_info)
        expected_views = keymap(partial(pd.Timestamp, tz='UTC'), {
            '2014-01-02': np.array([[10.0],
                                    [1.0]]),
            '2014-01-03': np.array([[11.0],
                                    [2.0]]),
        })

        with tmp_asset_finder(equities=simple_asset_info) as finder:
            expected_output = pd.DataFrame(
                list(concatv([10] * nassets, [11] * nassets)),
                index=pd.MultiIndex.from_product((
                    sorted(expected_views.keys()),
                    finder.retrieve_all(simple_asset_info.index),
                )),
                columns=('value',),
            )
            dates = self.dates
            self._run_pipeline(
                expr,
                deltas,
                None,
                expected_views,
                expected_output,
                finder,
                calendar=dates,
                start=dates[1],
                end=dates[-1],
                window_length=2,
                compute_fn=np.nanmax,
            )

    def test_deltas_before_index_0_macro(self):
        df = empty_dataframe(
            ('value', 'float64'),
            ('asof_date', 'datetime64[ns]'),
            ('timestamp', 'datetime64[ns]'),
        )
        expr = bz.data(df, name='expr', dshape=self.macro_dshape)

        T = pd.Timestamp
        # These data are interesting because we have four rows with an asof
        # date prior to the start of the query window. The first, second, and
        # fourth rows should become the best-known value on their timestamp.
        # The third row's asof date is less than the second row's asof date so,
        # due to forward filling rules, it is *not* the most recent value on
        # its timestamp. The value for row three should never be shown to the
        # user.
        deltas_df = pd.DataFrame({
            'value': [0.0, 1.0, 2.0, 3.0],
            'asof_date': [
                T('2013-12-01'),
                T('2013-12-15'),
                T('2013-12-02'),  # not more recent than the previous day
                T('2013-12-16'),
            ],
            'timestamp': [
                T('2014-01-01 23:00'),
                T('2014-01-02 23:00'),
                T('2014-01-03 23:00'),
                T('2014-01-04 23:00'),
            ],
        })
        deltas = bz.data(deltas_df, name='deltas', dshape=self.macro_dshape)

        expected_views = keymap(partial(pd.Timestamp, tz='UTC'), {
            '2014-01-02': np.array([[0.0],
                                    [0.0]]),
            '2014-01-03': np.array([[1.0],
                                    [1.0]]),
            # The third row's value of 2.0 is *not* the best known value
            # because its asof date of 2013-12-02 is earlier than the previous
            # row's asof date of 2013-12-15. We continue to surface the second
            # row's value on this day.
            '2014-01-04': np.array([[1.0],
                                    [1.0]]),
            '2014-01-05': np.array([[3.0],
                                    [3.0]]),
        })

        with tmp_asset_finder(equities=simple_asset_info) as finder:
            dates = pd.date_range('2014-01-01', '2014-01-05')
            self._run_pipeline(
                expr,
                deltas,
                None,
                expected_views,
                None,
                finder,
                calendar=dates,
                start=dates[1],
                end=dates[-1],
                window_length=2,
            )

    def test_deltas_on_same_ix_out_of_order_macro(self):
        df = empty_dataframe(
            ('value', 'float64'),
            ('asof_date', 'datetime64[ns]'),
            ('timestamp', 'datetime64[ns]'),
        )
        expr = bz.data(df, name='expr', dshape=self.macro_dshape)

        T = pd.Timestamp

        # These data are interesting because we have pairs of rows that come on
        # the same asof_date in index space. The catch is that the asof dates
        # are sometimes out of order relative to their timestamps. This is used
        # to test cases where we get novel rows for dates between trading days
        # (weekends and holidays) although we learn about them out of order.
        #
        # The first two rows both map to index 0 in the output. The first row
        # has an earlier timestamp but later asof date so it should be
        # selected.
        #
        # The third and fourth rows both map to index 1 in the output. The
        # fourth row (second in the group) has both a later timestamp and asof
        # date so it should be selected.
        #
        # The fifth and sixth rows both map to index 2 in the output. The fifth
        # row (first in the group) has an earlier timestamp but later asof date
        # so it should be selected.
        deltas_df = pd.DataFrame({
            'value': [
                0.0,  # selected
                1.0,  # ignored

                2.0,  # ignored
                3.0,  # selected

                4.0,  # selected
                5.0,  # ignored
            ],
            'asof_date': [
                # swapped order: second row is before the first
                T('2014-01-02'),
                T('2014-01-01'),

                # chronological order: second row is after the first
                T('2014-01-03'),
                T('2014-01-04'),

                # swapped order: second row is before the first
                T('2014-01-06'),
                T('2014-01-05'),
            ],
            'timestamp': [
                # we learn about all rows in monotonically increasing order
                T('2013-01-02 22:00'),
                T('2014-01-02 23:00'),
                T('2014-01-04 22:00'),
                T('2014-01-04 23:00'),
                T('2014-01-06 22:00'),
                T('2014-01-06 23:00'),
            ],
        })
        deltas = bz.data(deltas_df, name='deltas', dshape=self.macro_dshape)

        expected_views = keymap(partial(pd.Timestamp, tz='UTC'), {
            '2014-01-05': np.array([[0.0],
                                    [3.0]]),
            '2014-01-07': np.array([[3.0],
                                    [4.0]]),
        })

        with tmp_asset_finder(equities=simple_asset_info) as finder:
            # The dates queried are non-contiguous. We have two day groups to
            # capture the two day pairs in the input data.
            dates = pd.to_datetime(['2014-01-03', '2014-01-05', '2014-01-07'])
            self._run_pipeline(
                expr=expr,
                deltas=deltas,
                checkpoints=None,
                expected_views=expected_views,
                expected_output=None,
                finder=finder,
                calendar=dates,
                start=dates[1],
                end=dates[-1],
                window_length=2,
            )

    def test_stacked_deltas_macro(self):
        df = empty_dataframe(
            ('value', 'float64'),
            ('asof_date', 'datetime64[ns]'),
            ('timestamp', 'datetime64[ns]'),
        )
        expr = bz.data(df, name='expr', dshape=self.macro_dshape)

        T = pd.Timestamp

        # These data are interesting because they exercise the tie breaking of
        # adjustments. Here we have 4 rows which we learn about within a single
        # calendar index. The first row provides the most recently known value
        # for some day in the window. All of the following rows are adjustments
        # to the same (earlier) historical value. We expect that the first
        # row's value is the most recently know value, and the lookback window
        # will be filled with the *last* row's value. This is because each
        # adjustment gets applied in timestamp order, and the last row was
        # learned most recently.
        deltas_df = pd.DataFrame({
            'value': [
                0.0,  # selected
                1.0,  # ignored
                2.0,  # ignored
                3.0,  # ignored
                4.0,  # selected
            ],
            'asof_date': [
                # the first row is for current data
                T('2014-01-02'),

                # all other rows are restating the same historical value
                T('2013-12-01'),
                T('2013-12-01'),
                T('2013-12-01'),
                T('2013-12-01'),
            ],
            'timestamp': [
                # we learn about all rows within a single calendar index
                T('2014-01-02 23:00'),
                T('2014-01-02 23:01'),
                T('2014-01-02 23:02'),
                T('2014-01-02 23:03'),
                T('2014-01-02 23:04'),
            ],
        })
        deltas = bz.data(deltas_df, name='deltas', dshape=self.macro_dshape)

        expected_views = keymap(partial(pd.Timestamp, tz='UTC'), {
            '2014-01-03': np.array([[4.0],
                                    [4.0],
                                    [0.0]]),
        })

        with tmp_asset_finder(equities=simple_asset_info) as finder:
            # The dates queried are non-contiguous. We have two day groups to
            # capture the two day pairs in the input data.
            dates = pd.date_range('2014-01-01', '2014-01-03')
            self._run_pipeline(
                expr=expr,
                deltas=deltas,
                checkpoints=None,
                expected_views=expected_views,
                expected_output=None,
                finder=finder,
                calendar=dates,
                start=dates[-1],
                end=dates[-1],
                window_length=3,
            )

    @with_extra_sid()
    def test_novel_deltas(self, asset_info):
        base_dates = pd.DatetimeIndex([
            pd.Timestamp('2013-12-31'),
            pd.Timestamp('2014-01-03')
        ])
        repeated_dates = base_dates.repeat(3)
        baseline = pd.DataFrame({
            'sid': self.ASSET_FINDER_EQUITY_SIDS * 2,
            'value': (0., 1., 2., 1., 2., 3.),
            'int_value': (0, 1, 2, 1, 2, 3),
            'asof_date': repeated_dates,
            'timestamp': repeated_dates + pd.Timedelta(hours=23),
        })
        expr = bz.data(baseline, name='expr', dshape=self.dshape)
        deltas = bz.data(
            odo(
                bz.transform(
                    expr,
                    value=expr.value + 10,
                    timestamp=expr.timestamp + timedelta(days=1),
                ),
                pd.DataFrame,
            ),
            name='delta',
            dshape=self.dshape,
        )
        expected_views = keymap(partial(pd.Timestamp, tz='UTC'), {
            '2014-01-03': np.array([[10.0, 11.0, 12.0],
                                    [10.0, 11.0, 12.0],
                                    [10.0, 11.0, 12.0]]),
            '2014-01-06': np.array([[10.0, 11.0, 12.0],
                                    [10.0, 11.0, 12.0],
                                    [11.0, 12.0, 13.0]]),
        })

        if len(asset_info) == 4:
            def get_fourth_asset_view(expected_views, window_length):
                return valmap(
                    lambda view: np.c_[view, [np.nan] * window_length],
                    expected_views,
                )

            expected_views = get_fourth_asset_view(
                expected_views,
                window_length=3,
            )
            expected_output_buffer = [
                10,
                11,
                12,
                np.nan,
                11,
                12,
                13,
                np.nan,
            ]
        else:
            expected_output_buffer = [10, 11, 12, 11, 12, 13]

        cal = pd.DatetimeIndex([
            pd.Timestamp('2014-01-01'),
            pd.Timestamp('2014-01-02'),
            pd.Timestamp('2014-01-03'),
            # omitting the 4th and 5th to simulate a weekend
            pd.Timestamp('2014-01-06'),
        ])

        with tmp_asset_finder(equities=asset_info) as finder:
            expected_output = pd.DataFrame(
                expected_output_buffer,
                index=pd.MultiIndex.from_product((
                    sorted(expected_views.keys()),
                    finder.retrieve_all(asset_info.index),
                )),
                columns=('value',),
            )

            self._run_pipeline(
                expr,
                deltas,
                None,
                expected_views,
                expected_output,
                finder,
                calendar=cal,
                start=cal[2],
                end=cal[-1],
                window_length=3,
                compute_fn=op.itemgetter(-1),
            )

    def test_novel_deltas_macro(self):
        base_dates = pd.DatetimeIndex([
            pd.Timestamp('2013-12-31'),
            pd.Timestamp('2014-01-03')
        ])
        baseline = pd.DataFrame({
            'value': (0., 1.),
            'asof_date': base_dates,
            'timestamp': base_dates + pd.Timedelta(days=1),
        })
        expr = bz.data(baseline, name='expr', dshape=self.macro_dshape)
        deltas = bz.data(baseline, name='deltas', dshape=self.macro_dshape)
        deltas = bz.transform(
            deltas,
            value=deltas.value + 10,
            timestamp=deltas.timestamp + timedelta(days=1),
        )
        nassets = len(simple_asset_info)
        expected_views = keymap(partial(pd.Timestamp, tz='UTC'), {
            '2014-01-03': np.array([[10.0],
                                    [10.0],
                                    [10.0]]),
            '2014-01-06': np.array([[10.0],
                                    [10.0],
                                    [11.0]]),
        })

        cal = pd.DatetimeIndex([
            pd.Timestamp('2014-01-01'),
            pd.Timestamp('2014-01-02'),
            pd.Timestamp('2014-01-03'),
            # omitting the 4th and 5th to simulate a weekend
            pd.Timestamp('2014-01-06'),
        ])

        def get_expected_output(expected_views, values, asset_info):
            return pd.DataFrame(
                list(concatv(*([value] * nassets for value in values))),
                index=pd.MultiIndex.from_product(
                    (sorted(expected_views.keys()),
                     finder.retrieve_all(asset_info.index),)
                ), columns=('value',),
            )

        with tmp_asset_finder(equities=simple_asset_info) as finder:
            expected_output = get_expected_output(
                expected_views,
                [10, 11],
                simple_asset_info,
            )
            self._run_pipeline(
                expr,
                deltas,
                None,
                expected_views,
                expected_output,
                finder,
                calendar=cal,
                start=cal[2],
                end=cal[-1],
                window_length=3,
                compute_fn=op.itemgetter(-1),
            )

    test_checkpoints_dates = pd.date_range('2013-12-31', '2014-01-04')
    test_checkpoints_expected_view_date = pd.Timestamp('2014-01-03')

    def _test_checkpoints_macro(self, checkpoints, ffilled_value=-1.0):
        """Simple checkpoints test that accepts a checkpoints dataframe and
        the expected value for 2014-01-03 for macro datasets.

        The underlying data has value -1.0 on 2014-01-01 and 1.0 on 2014-01-04.

        Parameters
        ----------
        checkpoints : pd.DataFrame
            The checkpoints data.
        ffilled_value : float, optional
            The value to be read on the third, if not provided, it will be the
            value in the base data that will be naturally ffilled there.
        """
        dates = self.test_checkpoints_dates[[1, -1]]
        asof_dates = dates - pd.Timedelta(days=1)
        timestamps = asof_dates + pd.Timedelta(hours=23)
        baseline = pd.DataFrame({
            'value': [-1.0, 1.0],
            'asof_date': asof_dates,
            'timestamp': timestamps,
        })

        nassets = len(simple_asset_info)
        expected_views = keymap(lambda t: t.tz_localize('UTC'), {
            self.test_checkpoints_expected_view_date: (
                np.array([[ffilled_value]])
            ),
            self.test_checkpoints_dates[-1]: np.array([[1.0]]),
        })

        with tmp_asset_finder(equities=simple_asset_info) as finder:
            expected_output = pd.DataFrame(
                list(concatv([ffilled_value] * nassets, [1.0] * nassets)),
                index=pd.MultiIndex.from_product((
                    sorted(expected_views.keys()),
                    finder.retrieve_all(simple_asset_info.index),
                )),
                columns=('value',),
            )

            self._run_pipeline(
                bz.data(baseline, name='expr', dshape=self.macro_dshape),
                None,
                bz.data(
                    checkpoints,
                    name='expr_checkpoints',
                    dshape=self.macro_dshape,
                ),
                expected_views,
                expected_output,
                finder,
                calendar=pd.date_range('2014-01-01', '2014-01-04'),
                start=pd.Timestamp('2014-01-03'),
                end=dates[-1],
                window_length=1,
                compute_fn=op.itemgetter(-1),
            )

    @parameter_space(checkpoints_ts_fuzz_minutes=range(-5, 5))
    def test_checkpoints_macro(self, checkpoints_ts_fuzz_minutes):
        ffilled_value = 0.0

        checkpoints_ts = (
            self.test_checkpoints_expected_view_date -
            pd.Timedelta(days=1)
        )
        checkpoints = pd.DataFrame({
            'value': [ffilled_value],
            'asof_date': checkpoints_ts,
            'timestamp': (
                checkpoints_ts +
                # Fuzz the checkpoints timestamp a little so that it doesn't
                # align with the data query time. This should not affect the
                # correctness of the output.
                pd.Timedelta(minutes=checkpoints_ts_fuzz_minutes)
            ),
        })

        self._test_checkpoints_macro(checkpoints, ffilled_value)

    def test_empty_checkpoints_macro(self):
        empty_checkpoints = pd.DataFrame({
            'value': [],
            'asof_date': [],
            'timestamp': [],
        })

        self._test_checkpoints_macro(empty_checkpoints)

    def test_checkpoints_out_of_bounds_macro(self):
        # provide two checkpoints, one before the data in the base table
        # and one after, these should not affect the value on the third
        asof_dates = self.test_checkpoints_dates[[0, -1]]
        out_of_bounds = pd.DataFrame({
            'value': [-2, 2],
            'asof_date': asof_dates,
            'timestamp': asof_dates + pd.Timedelta(hours=23),
        })

        # Add a single checkpoint on the query day with a timestamp of exactly
        # the data query time. This should not get pulled to overwrite the
        # expected data on the 3rd.
        exact_query_time = pd.DataFrame({
            'value': [1],
            'asof_date': [
                self.test_checkpoints_expected_view_date -
                pd.Timedelta(days=1)
            ],
            'timestamp': [self.test_checkpoints_expected_view_date],
        })

        self._test_checkpoints_macro(
            pd.concat([out_of_bounds, exact_query_time]),
        )

    def _test_checkpoints(self, checkpoints, ffilled_values=None):
        """Simple checkpoints test that accepts a checkpoints dataframe and
        the expected value for 2014-01-03.

        The underlying data has value -(sid + 1) on 2014-01-01 and sid + 1 on
        2014-01-04.

        Parameters
        ----------
        checkpoints : pd.DataFrame
            The checkpoints data.
        ffilled_value : float, optional
            The value to be read on the third, if not provided, it will be the
            value in the base data that will be naturally ffilled there.
        """
        nassets = len(simple_asset_info)

        dates = self.test_checkpoints_dates[[1, -1]]

        asof_dates = dates - pd.Timedelta(days=1)
        asof_dates_repeated = np.tile(asof_dates, nassets)
        timestamps = asof_dates + pd.Timedelta(hours=23)
        timestamps_repeated = np.tile(timestamps, nassets)

        values = simple_asset_info.index.values + 1
        values = np.hstack((values[::-1], values))
        baseline = pd.DataFrame({
            'sid': np.tile(simple_asset_info.index, 2),
            'value': values,
            'asof_date': asof_dates_repeated,
            'timestamp': timestamps_repeated,
        })

        if ffilled_values is None:
            ffilled_values = baseline.value.iloc[:nassets]

        updated_values = baseline.value.iloc[nassets:]

        expected_views = keymap(partial(pd.Timestamp, tz='UTC'), {
            self.test_checkpoints_expected_view_date: [ffilled_values],
            self.test_checkpoints_dates[-1]: [updated_values],
        })

        with tmp_asset_finder(equities=simple_asset_info) as finder:
            expected_output = pd.DataFrame(
                list(concatv(ffilled_values, updated_values)),
                index=pd.MultiIndex.from_product((
                    sorted(expected_views.keys()),
                    finder.retrieve_all(simple_asset_info.index),
                )),
                columns=('value',),
            )

            self._run_pipeline(
                bz.data(baseline, name='expr', dshape=self.value_dshape),
                None,
                bz.data(
                    checkpoints,
                    name='expr_checkpoints',
                    dshape=self.value_dshape,
                ),
                expected_views,
                expected_output,
                finder,
                calendar=pd.date_range('2014-01-01', '2014-01-04'),
                start=pd.Timestamp('2014-01-03'),
                end=dates[-1],
                window_length=1,
                compute_fn=op.itemgetter(-1),
            )

    @parameter_space(checkpoints_ts_fuzz_minutes=range(-5, 5))
    def test_checkpoints(self, checkpoints_ts_fuzz_minutes):
        nassets = len(simple_asset_info)
        ffilled_values = (np.arange(nassets, dtype=np.float64) + 1) * 10
        dates = pd.Index([pd.Timestamp('2014-01-01')] * nassets)
        checkpoints = pd.DataFrame({
            'sid': simple_asset_info.index,
            'value': ffilled_values,
            'asof_date': dates,
            'timestamp': (
                dates +
                # Fuzz the checkpoints timestamp a little so that it doesn't
                # align with the data query time. This should not affect the
                # correctness of the output.
                pd.Timedelta(days=1, minutes=checkpoints_ts_fuzz_minutes)
            ),
        })
        self._test_checkpoints(checkpoints, ffilled_values)

    def test_empty_checkpoints(self):
        checkpoints = pd.DataFrame({
            'sid': [],
            'value': [],
            'asof_date': [],
            'timestamp': [],
        })

        self._test_checkpoints(checkpoints)

    def test_checkpoints_out_of_bounds(self):
        nassets = len(simple_asset_info)
        # provide two sets of checkpoints, one before the data in the base
        # table and one after, these should not affect the value on the third
        asof_dates = self.test_checkpoints_dates[[0, -1]]
        asof_dates_repeated = np.tile(asof_dates, nassets)
        ffilled_values = (np.arange(nassets) + 2) * 10
        ffilled_values = np.hstack((ffilled_values[::-1], ffilled_values))
        out_of_bounds = pd.DataFrame({
            'sid': np.tile(simple_asset_info.index, 2),
            'value': ffilled_values,
            'asof_date': asof_dates_repeated,
            'timestamp': asof_dates_repeated + pd.Timedelta(hours=23),
        })

        # Add a single checkpoint on the query day with a timestamp of exactly
        # the data query time. This should not get pulled to overwrite the
        # expected data on the 3rd.
        exact_query_time = pd.DataFrame({
            'sid': simple_asset_info.index,
            'value': simple_asset_info.index + 1,
            'asof_date': (
                self.test_checkpoints_expected_view_date -
                pd.Timedelta(days=1)
            ),
            'timestamp': self.test_checkpoints_expected_view_date,
        })

        self._test_checkpoints(pd.concat([out_of_bounds, exact_query_time]))

    def test_id_take_last_in_group_sorted(self):
        """
        input
        asof_date     timestamp     other  value
        2014-01-03    2014-01-04 00     3      3
        2014-01-02    2014-01-04 00     2      2

        output (expected):

                    other  value
        2014-01-02    NaN    NaN
        2014-01-03    NaN    NaN
        2014-01-06      3      3
        """

        dates = pd.DatetimeIndex([
            pd.Timestamp('2014-01-02'),
            pd.Timestamp('2014-01-03'),
            pd.Timestamp('2014-01-06'),
        ]).tz_localize('UTC')

        T = pd.Timestamp
        df = pd.DataFrame(
            columns=['asof_date', 'timestamp', 'other', 'value'],
            data=[
                # asof-dates are flipped in terms of order so that if we
                # don't sort on asof-date before getting the last in group,
                # we will get the wrong result.
                [T('2014-01-03'), T('2014-01-04 00'), 3, 3],
                [T('2014-01-02'), T('2014-01-04 00'), 2, 2],
            ],
        )
        fields = OrderedDict(self.macro_dshape.measure.fields)
        fields['other'] = fields['value']
        expected = pd.DataFrame(
            data=[[np.nan, np.nan],   # 2014-01-02
                  [np.nan, np.nan],   # 2014-01-03
                  [3,      3]],       # 2014-01-06
            columns=['other', 'value'],
            index=dates,
        )
        self._test_id_macro(
            df,
            var * Record(fields),
            expected,
            self.asset_finder,
            ('other', 'value'),
            dates=dates,
        )


class MiscTestCase(ZiplineTestCase):
    def test_exprdata_repr(self):
        strd = set()

        class BadRepr(object):
            """A class which cannot be repr'd.
            """
            def __init__(self, name):
                self._name = name

            def __repr__(self):  # pragma: no cover
                raise AssertionError('ayy')

            def __str__(self):
                strd.add(self)
                return self._name

        assert_equal(
            repr(ExprData(
                expr=BadRepr('expr'),
                deltas=BadRepr('deltas'),
                checkpoints=BadRepr('checkpoints'),
                odo_kwargs={'a': 'b'},
            )),
            "ExprData(expr=expr, deltas=deltas,"
            " checkpoints=checkpoints, odo_kwargs={'a': 'b'})",
        )

    def test_exprdata_eq(self):
        dshape = 'var * {sid: int64, asof_date: datetime, value: float64}'
        base_expr = bz.symbol('base', dshape)
        checkpoints_expr = bz.symbol('checkpoints', dshape)

        # use a nested dict to emulate real call sites
        odo_kwargs = {'a': {'c': 1, 'd': 2}, 'b': {'e': 3, 'f': 4}}

        actual = ExprData(
            expr=base_expr,
            deltas=None,
            checkpoints=checkpoints_expr,
            odo_kwargs=odo_kwargs,
        )
        same = ExprData(
            expr=base_expr,
            deltas=None,
            checkpoints=checkpoints_expr,
            odo_kwargs=odo_kwargs,
        )
        self.assertEqual(actual, same)
        self.assertEqual(hash(actual), hash(same))

        different_obs = [
            actual.replace(expr=bz.symbol('not base', dshape)),
            actual.replace(expr=bz.symbol('not deltas', dshape)),
            actual.replace(checkpoints=bz.symbol('not checkpoints', dshape)),
            actual.replace(checkpoints=None),
            actual.replace(odo_kwargs={
                # invert the leaf values
                ok: {ik: ~iv for ik, iv in ov.items()}
                for ok, ov in odo_kwargs.items()
            }),
        ]

        for different in different_obs:
            self.assertNotEqual(actual, different)

        actual_with_none_odo_kwargs = actual.replace(odo_kwargs=None)
        same_with_none_odo_kwargs = same.replace(odo_kwargs=None)

        self.assertEqual(
            actual_with_none_odo_kwargs,
            same_with_none_odo_kwargs,
        )
        self.assertEqual(
            hash(actual_with_none_odo_kwargs),
            hash(same_with_none_odo_kwargs),
        )

    def test_blaze_loader_lookup_failure(self):
        class D(DataSet):
            c = Column(dtype='float64')

        with self.assertRaises(KeyError) as e:
            BlazeLoader()(D.c)
        assert_equal(str(e.exception), 'D.c::float64')



from unittest import TestCase

from pandas import Timestamp, Timedelta

from zipline.utils.cache import CachedObject, Expired, ExpiringCache


class CachedObjectTestCase(TestCase):

    def test_cached_object(self):
        expiry = Timestamp('2014')
        before = expiry - Timedelta('1 minute')
        after = expiry + Timedelta('1 minute')

        obj = CachedObject(1, expiry)

        self.assertEqual(obj.unwrap(before), 1)
        self.assertEqual(obj.unwrap(expiry), 1)  # Unwrap on expiry is allowed.
        with self.assertRaises(Expired) as e:
            obj.unwrap(after)
        self.assertEqual(e.exception.args, (expiry,))

    def test_expired(self):
        always_expired = CachedObject.expired()

        for dt in Timestamp.min, Timestamp.now(), Timestamp.max:
            with self.assertRaises(Expired):
                always_expired.unwrap(dt)


class ExpiringCacheTestCase(TestCase):

    def test_expiring_cache(self):
        expiry_1 = Timestamp('2014')
        before_1 = expiry_1 - Timedelta('1 minute')
        after_1 = expiry_1 + Timedelta('1 minute')

        expiry_2 = Timestamp('2015')
        after_2 = expiry_1 + Timedelta('1 minute')

        expiry_3 = Timestamp('2016')

        cache = ExpiringCache()

        cache.set('foo', 1, expiry_1)
        cache.set('bar', 2, expiry_2)

        self.assertEqual(cache.get('foo', before_1), 1)
        # Unwrap on expiry is allowed.
        self.assertEqual(cache.get('foo', expiry_1), 1)

        with self.assertRaises(KeyError) as e:
            self.assertEqual(cache.get('foo', after_1))
        self.assertEqual(e.exception.args, ('foo',))

        # Should raise same KeyError after deletion.
        with self.assertRaises(KeyError) as e:
            self.assertEqual(cache.get('foo', before_1))
        self.assertEqual(e.exception.args, ('foo',))

        # Second value should still exist.
        self.assertEqual(cache.get('bar', after_2), 2)

        # Should raise similar KeyError on non-existent key.
        with self.assertRaises(KeyError) as e:
            self.assertEqual(cache.get('baz', expiry_3))
        self.assertEqual(e.exception.args, ('baz',))
